from math import floor, ceil

from scipy import stats
import numpy as np

from cheno.util import _gen_nonparam

class Friedman(object):

    def solve(self, p, w, k, n=None, epoch=1000):
        f = stats.chi2.ppf(1 - p, k - 1)
        n_prime = int(round((f / w) / (k - 1)))

        if n is None:
            n = n_prime
        elif n != n_prime:
            raise ValueError("Constrained size value, n, incorrect.")

        _ranks = self.bpass(f, k, n, epoch)
        return _ranks.T, self._gen_data(_ranks, k)

    @staticmethod
    def bpass(f, k, n, epoch):
        f_prime = ((f + 3 * n * (k + 1)) / 12) * (n * k * (k + 1))
        ranks = np.array([list(range(1, k + 1)) for _ in range(n)])

        max_ss = sum([(n * x) ** 2 for x in range(1, k + 1)])
        avg_ss = sum([(n * x) for x in range(1, k + 1)]) / k
        min_ss = []
        for i in range(1, k + 1):
            if i % 2 == 0:
                min_ss.append(ceil(avg_ss))
            else:
                min_ss.append(floor(avg_ss))
        min_ss = sum([x ** 2 for x in min_ss])
        if not min_ss <= f_prime <= max_ss:
            raise ValueError(
                "Impossible test statistics. Please try a different f-value.")

        def obj(x):
            s = [sum(x[:, i]) ** 2 for i in range(k)]
            return sum(s) - f_prime, sum(s), s

        min_d, s, s_a = obj(ranks)
        best_ranks = ranks

        for i in range(epoch):
            m_ranks = np.copy(best_ranks)
            m_ranks[i % n] = np.random.permutation(m_ranks[i % n])
            d, s1, s_a1 = obj(m_ranks)

            if abs(d) < abs(min_d):
                best_ranks = m_ranks
                min_d = d

        return best_ranks

    @staticmethod
    def _gen_data(ranks, k):
        values = np.empty(ranks.shape)
        for i, x in enumerate(ranks):
            values[i] = (_gen_nonparam([x], k))[0]
        return values.T


if __name__ == '__main__':
    # Example
    p = 0.9232628347180212
    w = 0.0016000000000000606

    # Commands
    f = Friedman()
    _, data = f.solve(p, w, 4)

    # Results
    _, p_prime = stats.friedmanchisquare(*data)
    print(p_prime)
