from cheno.util import count_none, _gen_nonparam, rand_generate_subset

from math import floor, ceil, sqrt

from scipy import stats

class MannWhitney(object):

    def __init__(self, low_n=10, high_n=200, throw_error=True):
        self.low_n = low_n
        self.high_n = high_n
        self.throw_error = throw_error

    def solve(self, p, r, ns=None, tails=1):
        ns = [None, None] if ns is None else ns

        if len(ns) != 2:
            if self.throw_error:
                raise ValueError("Incorrect number of provided sample sizes.")
            else:
                ns = [None, None]

        if tails != 1 and tails != 2:
            if self.throw_error:
                raise ValueError("Invalid number of tails. NHST accept either one tail or two tails.")
            else:
                tails = 2

        z = stats.norm.ppf(p / tails)
        n_prime = int(round(((z / r) ** 2)))

        if not self.low_n <= n_prime / 2 <= self.high_n:
            if self.throw_error:
                raise ValueError(f"A p-value of {p} with {tails}-tails implies a sample size sum of {n_prime},"
                                 f"which would produce sample sizes outside of the bounds. Loosen the bounds or"
                                 f"change the inference statistics.")

        if count_none(ns) == 0:
            if sum(ns) != n_prime:
                if self.throw_error:
                    raise ValueError(f"Incorrect sample sizes. Sample size sum has to be equal to {n_prime}.")
        elif count_none(ns) == 1:
            prov_n_idx, prov_n = ([(i, n) for i, n in enumerate(ns) if n is not None])[0]
            ns[1 - prov_n_idx] = n_prime - prov_n
        else:
            ns[0] = ceil(n_prime / 2)
            ns[1] = floor(n_prime / 2)

        ranks = MannWhitney.bpass(z, ns)

        return ranks, _gen_nonparam(ranks, n_prime)

    @staticmethod
    def bpass(z, ns):

        w_mean = min(ns[0], ns[1]) * (ns[0] + ns[1] + 1) / 2
        se_w = sqrt(ns[0] * ns[1] * (ns[0] + ns[1] + 1) / 12)
        w = ceil(z * se_w + w_mean)

        if ns[0] > ns[1]:
            return list(reversed(rand_generate_subset(ns[1], ns[0], w)))
        else:
            return rand_generate_subset(ns[0], ns[1], w)

if __name__ == "__main__":
    # Example
    p = 0.4210690066046685
    r = -0.05689154982036873

    # Commands
    mw = MannWhitney()
    _, data = mw.solve(p, r, tails=2)

    # Results
    _, p_prime = stats.mannwhitneyu(data[0], data[1])
    print(p_prime)