from scipy import stats
from numpy import isclose


class rANOVA(object):

    b_size = None
    b_means = []
    b_stds = []
    B_SIZE_DEFAULT = (5, 100)
    B_MEANS_DEFAULT = (0, 7)
    B_STDS_DEFAULT = (.1, 2)

    def __init__(self, groups=4, throw_error=True):
        self.throw_error = throw_error
        self.groups = groups
        self._init_bounds()

    def _init_bounds(self):
        self.b_means = [self.B_MEANS_DEFAULT for _ in range(self.groups)]
        self.b_stds = [self.B_STDS_DEFAULT for _ in range(self.groups)]
        self.b_size = self.B_SIZE_DEFAULT

    def set_bounds(self, b_means=[], b_stds=[], b_size=None):
        if len(b_means) == 0:
            pass
        elif len(b_means) == self.groups:
            self.b_means = b_means
        elif self.throw_error:
            raise ValueError(
                f"Incorrect means bounds array length. Expected length {self.groups}, but got {len(b_means)}.")

        if len(b_stds) == 0:
            pass
        elif len(b_stds) == self.groups:
            self.b_stds = b_stds
        elif self.throw_error:
            raise ValueError(
                f"Incorrect stds bounds array length. Expected length {self.groups}, but got {len(b_stds)}.")

        if b_size is None:
            pass
        elif len(b_size) == 2:
            self.b_size = b_size
        elif self.throw_error:
            raise ValueError(
                f"Incorrect size bounds tuple length. Expected length 2, but got {len(b_size)}.")

    def solve(self, p, eta2, means=None, stds=None, size=None, g_mean=0):
        # Initialization
        size = None if size is None else size
        means = [None for _ in range(self.groups)] if means is None else means
        stds = [None for _ in range(self.groups)] if stds is None else stds

        size = self._find_size(p, eta2, size)
        means, stds = self._find_means_sds(means, stds, size, g_mean)

        return means, stds, size

    def _find_size(self, p, eta2, size):
        df_M = self.groups - 1
        n = None
        min_diff = float("inf")
        for i in range(self.b_size[0], self.b_size[1] + 1):
            df_R = (self.groups - 1) * (i - 1)
            f = stats.f.ppf(1 - p, df_M, df_R)
            cur_min_diff = abs((f * (i - 1)) - (eta2 / (1 - eta2)))
            if cur_min_diff < min_diff:
                min_diff = cur_min_diff
                n = i

        if (size is not None and
            not isclose(n, size, atol=5) and
            not self.b_size[0] <= n <= self.b_size[1] and
                self.throw_error):
            raise ValueError(
                "User provided sample size doesn't fit the inferential statistics.")

        return n

    def _find_means_sds(self, means, stds, size, g_mean):
        return means, stds


if __name__ == "__main__":
    # Example (Andy Fields Book)
    F = 3.79
    df_M = 3
    df_R = 21
    p = 1 - stats.f.cdf(F, df_M, df_R)
    eta2 = 0.791488146
    
    # User provided constraints
    size = 5

    # Commands
    r = rANOVA()
    means, stds, size = r.solve(p, eta2, size=size)

    # Results
    print(f"Means: {means}")
    print(f"Stds: {stds}")
    print(f"Size: {size}\n")
    
