from cheno.util import rand_generate_subset, _gen_nonparam

from math import sqrt
from random import shuffle

from scipy import stats
import numpy as np

class Wilcoxon(object):

    def __init__(self, sample_size_lb=10, sample_size_ub=200, throw_error=True):
        self.lb = sample_size_lb
        self.ub = sample_size_ub
        self.throw_error = throw_error

    def solve(self, p, r, n=None, tails=1):
        if tails != 1 and tails != 2:
            if self.throw_error:
                raise ValueError("Invalid number of tails. NHST accept either one tail or two tails.")
            else:
                tails = 2

        z = stats.norm.ppf(p / tails)

        with np.errstate(divide='ignore', invalid='ignore'):
            n_prime = int(round(((z / r) ** 2)))

        if n_prime % 2 != 0:
            n_prime -= 1

        if not self.lb <= n_prime / 2 <= self.ub:
            if self.throw_error:
                raise ValueError(f"A p-value of {p} with {tails}-tails implies a sample size sum of {n_prime},"
                                 f"which would produce sample sizes outside of the bounds. Loosen the bounds or"
                                 f"change the inference statistics.")

        if n is None:
            n = n_prime / 2
        elif 2 * n != n_prime:
            if self.throw_error:
                raise ValueError("Provided sample size will not produce desired inference statistics. "
                                 "Pick a different sample size or remove it altogether.")

        ranks = Wilcoxon.bpass(z, n)
        diffs = _gen_nonparam(ranks, n_prime, low=0, high=10)
        return ranks, self._gen_data(diffs)

    @staticmethod
    def bpass(z, n):
        t_mean = n * (n + 1) / 4
        se_t = sqrt(n * (n + 1) * (2 * n + 1) / 24)
        t = round(z * se_t + t_mean)

        if t < 0:
            raise ValueError("Negative test-statistics")

        ns = list(range(int(round(n / 2)) + 1))
        shuffle(ns)
        for i in ns:
            if i * (i + 1) / 2 <= t <= i * (n + (n - i + 1)) / 2:
                return rand_generate_subset(i, int(n - i), int(t))

    @staticmethod
    def _gen_data(diffs):
        n1 = len(diffs[0])
        n2 = len(diffs[1])

        data1 = np.random.uniform(-10, 10, n1 + n2)
        data2 = np.empty(n1 + n2)
        for i, x in enumerate(data1):
            if i < n1:
                data2[i] = x + diffs[0][i]
            else:
                data2[i] = x - diffs[1][i - n1]

        return [data1, data2]

if __name__ == "__main__":
    pass

