from math import ceil, floor, isclose, sqrt
from random import randint, uniform

from scipy import stats, optimize

from cheno.util import _gen_param, count_non_null, count_none


class ANOVA(object):

    def __init__(self, k, m_bounds=None, s_bounds=None, n_bounds=None, throw_error=True):
        self.k = k
        self.throw_error = throw_error

        if m_bounds is None:
            self.m_bounds = [[0, 7] for _ in range(k)]
        if s_bounds is None:
            self.s_bounds = [[.1, 2] for _ in range(k)]
        if n_bounds is None:
            self.n_bounds = [[5, 100] for _ in range(k)]

    def solve(self, p, eta2, ns=None, means=None, sds=None, g_mean=0):
        ns, means, sds = self._init_parameters(ns, means, sds)

        ns = self._find_sample_szs(p, eta2, ns)

        sds = self._find_sds(eta2, ns, means, sds)

        means = self._find_means(eta2, ns, means, sds, g_mean)

        return means, sds, ns, _gen_param(means, sds, ns)

    def _init_parameters(self, ns, means, sds):
        # Initialization
        ns = [None for _ in range(self.k)] if ns is None else ns
        means = [None for _ in range(self.k)] if means is None else means
        sds = [None for _ in range(self.k)] if sds is None else sds

        # General check on the size of the constraints
        if len(ns) != self.k:
            if self.throw_error:
                raise ValueError(
                    "Provided number of sample size doesn't match number of groups.")
            else:
                ns = [None for _ in range(self.k)]
        if len(means) != self.k:
            if self.throw_error:
                raise ValueError(
                    "Provided number of means doesn't match number of groups")
            else:
                means = [None for _ in range(self.k)]
        if len(sds) != self.k:
            if self.throw_error:
                raise ValueError(
                    "Provided number of standard deviations doesn't match number of groups")
            else:
                sds = [None for _ in range(self.k)]

        return ns, means, sds

    def _find_sample_szs(self, p, eta2, ns):
        is_in_range = [self.n_bounds[i][0] <= n <= self.n_bounds[i][1]
                       for i, n in enumerate(ns) if n is not None]
        if not all(is_in_range):
            if self.throw_error:
                raise ValueError("Provided sample sizes are out of bounds. Either loosen the bounds or"
                                 " change the sample sizes.")

        low = sum(map(lambda x: x[0], self.n_bounds))
        high = sum(map(lambda x: x[1], self.n_bounds))

        dfb = self.k - 1
        f_from_eta2 = abs(eta2 / (1 - eta2))
        min_diff = float("inf")
        n_total = None
        for i in range(low, high + 1):
            dfw = (i - 1) - dfb
            f_from_p = stats.f.ppf(1 - p, dfb, dfw) * (dfb / dfw)
            cur_diff = abs(f_from_p - f_from_eta2)
            if cur_diff < min_diff:
                min_diff = cur_diff
                n_total = i
            if cur_diff == 0:
                n_total = i
                break

        if count_none(ns) == 0:
            if sum(ns) != n_total:
                if self.throw_error:
                    raise ValueError("Provided sample size/s will not produce desired inference statistics. "
                                     "Pick different sample size/s or remove it/them altogether.")
        else:
            factor = (
                n_total - sum([n for n in ns if n is not None])) / count_none(ns)
            while count_none(ns) != 1:
                idx = ([i for i, n in enumerate(ns) if n is None])[0]
                ns[idx] = floor(factor)
            idx = ([i for i, n in enumerate(ns) if n is None])[0]
            ns[idx] = ceil(factor)

        return ns

    def _find_sds(self, eta2, ns, means, sds):
        is_in_range = [self.s_bounds[i][0] <= sd <= self.s_bounds[i][1]
                       for i, sd in enumerate(sds) if sd is not None]
        if not all(is_in_range):
            if self.throw_error:
                raise ValueError("Provided standard deviations are out of bounds. Either loosen the bounds or"
                                 " change the standard deviations.")

        is_in_range = [self.m_bounds[i][0] <= m <= self.m_bounds[i][1]
                       for i, m in enumerate(means) if m is not None]
        if not all(is_in_range):
            if self.throw_error:
                raise ValueError("Provided means are out of bounds. Either loosen the bounds or"
                                 " change the means.")

        n_total = sum(ns)

        if count_none(means) == 0:
            g_mean = sum([n * m for n, m in zip(ns, means)]) / n_total
            ss_m = sum([n * m ** 2 for n, m in zip(ns, means)]) - \
                n_total * g_mean ** 2
            ss_r = ss_m * (1 - eta2) / eta2

            while count_none(sds) != 0:
                rest = ss_r - sum([(n - 1) * sd ** 2 for n,
                                   sd in zip(ns, sds) if sd is not None])
                df = sum([(n - 1) for n, sd in zip(ns, sds) if sd is None])
                factor = rest / df
                idx = [i for i, sd in enumerate(sds) if sd is None][0]
                sds[idx] = sqrt(factor * uniform(.9, 1.1)
                                ) if count_none(sds) != 1 else sqrt(factor)
        else:
            if count_none(sds) == self.k:
                idx = randint(0, self.k - 1)
                sds[idx] = uniform(self.s_bounds[idx][0],
                                   self.s_bounds[idx][1])
            while count_none(sds) != 0:
                factor = sum([n for n in sds if n is not None]) / \
                    count_non_null(sds)
                idx = ([i for i, n in enumerate(sds) if n is None])[0]
                sds[idx] = factor * uniform(.9, 1.1)

        return sds

    def _find_means(self, eta2, ns, means, sds, g_mean):
        n_total = sum(ns)

        if count_none(means) != 0:
            ss_r = sum([(n - 1) * (sd ** 2) for n, sd in zip(ns, sds)])
            ss_m = ss_r * eta2 / (1 - eta2)

            while count_none(means) > 2:
                rest = n_total * g_mean - \
                    sum([n * mean for n, mean in zip(ns, means) if mean is not None])
                df = sum([ns[i]
                          for i, mean in enumerate(means) if mean is None])
                factor = rest / df
                idx = [i for i, mean in enumerate(means) if mean is None][0]
                means[idx] = factor * uniform(.9, 1.1)

            if count_none(means) == 1:
                idx = ([i for i, mean in enumerate(means) if mean is None])[0]
                means[idx] = n_total * g_mean - \
                    sum([n * mean for n, mean in zip(ns, means) if mean is not None])
                ss_m_prime = sum(
                    [n * (mean - g_mean) ** 2 for n, mean in zip(ns, means) if mean is not None])

                if not isclose(ss_m_prime, ss_m):
                    if self.throw_error:
                        raise ValueError("Provided means will not produce desired inference statistics. "
                                         "Pick different means or remove them altogether.")
                    else:
                        return means
                else:
                    return means

            r1 = n_total * g_mean - \
                sum([n * m for n, m in zip(ns, means) if m is not None])
            r2 = ss_m + n_total * g_mean ** 2 - sum(
                [n * m ** 2 for n, m in zip(ns, means) if m is not None])
            coef = [n for n, m in zip(ns, means) if m is None]

            def obj(x):
                s1 = sum([c * v for c, v in zip(coef, x)])
                s2 = sum([c * v ** 2 for c, v in zip(coef, x)])
                return [s1 - r1, s2 - r2]

            res = optimize.fsolve(obj, [g_mean, g_mean])
            ids = [i for i, mean in enumerate(means) if mean is None]
            means[ids[0]] = res[0]
            means[ids[1]] = res[1]

        return means

    def get_const(self):
        return {"means": self.m_bounds, "stds": self.s_bounds, "sizes": self.n_bounds}
