from cheno.util import count_none, _gen_nonparam

from random import choice

from scipy import stats
import numpy as np

class Kruskal(object):
    sz_b = []
    k = None

    def __init__(self, bounds, throw_error=True):
        self.throw_error = throw_error
        if bounds is not None:
            self.k = len(bounds)
            for i, x in enumerate(bounds):
                self.sz_b.append(x)

    def solve(self, p, eta2, ns=None, pop=1000, epoch=1000):
        ns = self._find_sample_szs(p, eta2, ns)
        ranks = self._find_ranks(p, ns, pop, epoch)
        return ranks, _gen_nonparam(ranks, sum(ns))

    def _find_sample_szs(self, p, eta2, ns):
        if ns is None:
            ns = [None for _ in range(self.k)]
        if len(ns) != self.k:
            if self.throw_error:
                raise ValueError("Provided number of sample size doesn't match number of groups.")
            else:
                ns = [None for _ in range(self.k)]

        h = stats.chi2.ppf(1 - p, self.k - 1)
        n_total = int(round(((h - self.k + 1) / eta2) + self.k))

        if count_none(ns) == 0:
            if sum(ns) != n_total:
                if self.throw_error:
                    raise ValueError("Provided sample size will not produce desired inference statistics. "
                                     "Pick a different sample size or remove it altogether.")

        if not all([self.sz_b[i][0] <= n <= self.sz_b[i][1] for i, n in enumerate(ns) if n is not None]):
            if self.throw_error:
                raise ValueError(f"Provided sample sizes outside of the bounds. Loosen the bounds or"
                                 f"change the inference statistics.")

        for i, idx in enumerate([idx for idx, n in enumerate(ns) if n is None]):
            factor = (n_total - sum([n for n in ns if n is not None])) / count_none(ns)
            ns[idx] = round(factor)

        return ns

    def _find_ranks(self, p, ns, pop, epoch):
        # Init
        h = stats.chi2.ppf(1 - p, self.k - 1)
        n_total = sum(ns)
        h_prime = ((h + 3 * (n_total + 1)) / 12) * (n_total * (n_total + 1))
        best_rank_ls = []
        min_diff = float("inf")

        # Computation
        ls = self._gen_random_ranks_list(ns, pop=pop, h_prime=h_prime)

        cur_rank_ls = np.array(ls)

        mean_sq_sums = [s.sum() ** 2 / s.size for s in cur_rank_ls]
        cur_diff = h_prime - sum(mean_sq_sums)

        for _ in range(epoch):
            if cur_diff > 0:
                self._redistribute(cur_rank_ls, direction="inc")
            elif cur_diff < 0:
                self._redistribute(cur_rank_ls, direction="dec")
            else:
                best_rank_ls = cur_rank_ls
                break

            mean_sq_sums = [s.sum() ** 2 / s.size for s in cur_rank_ls]
            cur_diff = h_prime - sum(mean_sq_sums)

            if abs(cur_diff) < min_diff:
                best_rank_ls = cur_rank_ls

        return best_rank_ls

    def _gen_random_ranks_list(self, ns, pop=1, h_prime=0):
        min_diff = float("inf")
        best_rank_list = []
        for _ in range(pop):
            rank_list = []
            ranks = np.random.permutation(sum(ns)) + 1
            chunk_szs = [0] + ns
            cur_idx = 0

            for i in range(self.k):
                cur_chunk = ranks[cur_idx: cur_idx + chunk_szs[i + 1]]
                cur_chunk.sort()
                rank_list.append(cur_chunk)
                cur_idx += chunk_szs[i + 1]

            cur_rank_ls = np.array(rank_list)

            mean_sq_sums = [sum(s) ** 2 / len(s) for s in cur_rank_ls]
            cur_diff = abs(h_prime - sum(mean_sq_sums))
            if cur_diff < min_diff:
                min_diff = cur_diff
                best_rank_list = rank_list

        return best_rank_list

    def _redistribute(self, rank_ls, direction):
        indexed_mean_sq_sums = [(s.sum() ** 2 / s.size, i) for i, s in enumerate(rank_ls)]
        indexed_mean_sq_sums.sort(key=lambda x: x[0])

        if direction == "inc":
            send_idx = indexed_mean_sq_sums[0][1]
            receive_idx = choice([i for v, i in indexed_mean_sq_sums if v > indexed_mean_sq_sums[0][0]])
        elif direction == "dec":
            send_idx = indexed_mean_sq_sums[-1][1]
            receive_idx = choice([i for v, i in indexed_mean_sq_sums if v < indexed_mean_sq_sums[-1][0]])

        return self._swap(rank_ls, send_idx, receive_idx, direction)

    def _swap(self, rank_ls, send_idx, receive_idx, direction):
        for i, elem in enumerate(rank_ls[send_idx]):
            if direction == "inc":
                swap_idx, swap_elem = self._next_lesser(rank_ls[receive_idx], elem)
            elif direction == "dec":
                swap_idx, swap_elem = self._next_lesser(rank_ls[receive_idx], elem)

            if swap_idx == -1:
                continue
            else:
                rank_ls[send_idx][i] = swap_elem
                rank_ls[receive_idx][swap_idx] = elem

                (rank_ls[send_idx]).sort()

        return rank_ls

    @staticmethod
    def _next_lesser(ls, v):
        if ls is None or len(ls) <= 0:
            raise ValueError('Comparing values in an empty list.')
        # if not all(ls[i] <= ls[i + 1] for i in range(len(ls) - 1)):
        #     raise ValueError('List is not sorted.')

        for i, e in enumerate(reversed(ls)):
            if v > e:
                return len(ls) - i - 1, e
        return -1, -1


if __name__ == "__main__":
    # Example
    p = 0.5696958703043062
    eta2 = -0.0024922164286257007

    # Commands
    kw = Kruskal(None, None, None, None)
    _, data = kw.solve(p, eta2)

    # Results
    _, p_prime = stats.kruskal(*data)
    print(p_prime)