from cheno.util import count_non_null, count_none, _gen_param

from abc import ABC, abstractmethod
from json import load
from os import listdir

from random import uniform, randint
from itertools import product
from math import sqrt, isclose, floor, ceil
import multiprocessing as mp

from scipy import stats, optimize
from numpy import errstate, where, isnan, empty, copy

print(listdir())

with open('cheno/config.json') as json_data_file:
    defaults = load(json_data_file)


class Parametric(ABC):

    test = None
    bounds = None
    constraints = None
    throw_err = True

    def __init__(self, test, throw_err=True):
        self.test = test
        self.throw_err = throw_err
        self.bounds = defaults[test]['bounds']

    def setbounds(self, bounds):
        if Parametric.iscorrect_bounds(bounds):
            self.bounds = bounds

    def setconstraints(self, constraints):
        if Parametric.iscorrect_constraints(constraints):
            self.constraints = constraints
        else:
            self.constraints = defaults[self.test]['defaultConstraints']

    @abstractmethod
    def iscorrect_bounds(bounds):
        pass

    @abstractmethod
    def iscorrect_constraints(constraints):
        pass

    def solve(self, p, ef, tails=2, bounds=None, constraints=None):
        self.setbounds(bounds)
        self.setconstraints(constraints)
        m, s, n = self.inference_2_descriptive(p, ef, tails)
        data = self.descriptive_2_data(m, s, n)
        return m, s, n, data

    @abstractmethod
    def inference_2_descriptive(self, p, ef, tails):
        pass

    @abstractmethod
    def descriptive_2_data(self, m, s, n):
        pass


class StudentD_refactored(Parametric):

    def __init__(self, throw_err=True):
        super().__init__("StudentD", throw_err=throw_err)

    def iscorrect_bounds(bounds):
        if bounds is None:
            return False

    def iscorrect_constraints(constraints):
        if constraints is None:
            return False

    def descriptive_2_data(self, m, s, n):
        return _gen_param([m], [s], [n])

    def inference_2_descriptive(self, p, ef, tails):
        n = self._find_n(p, ef, 2)
        m, s = self._find_m_s(p, ef, tails, n)
        return m, s, n

    def _find_n(self, p, d, tails):
        n = None
        min_diff = float("inf")

        for i in range(self.bounds['n'][0], self.bounds['n'][1] + 1):
            t_from_p = abs(stats.t.ppf(1 - (p / tails), i - 1))
            t_from_d = abs(d * sqrt(i))
            cur_diff = abs(t_from_p - t_from_d)
            if cur_diff < min_diff:
                min_diff = cur_diff
                n = i

        if self.throw_err and self.constraints['n'] is not None and self.constraints['n'] != n:
            raise ValueError("Provided sample size will not produce desired inference statistics. "
                             "Pick a different sample size or remove it altogether.")

        return n

    def _find_m_s(self, p, ef, tails, n):
        m = self.constraints['m']
        s = self.constraints['s']
        t = abs(stats.t.ppf(1 - (p / tails), n - 1))
        d = abs(t / sqrt(n))

        self._corr_const(d)

        if self.constraints['s'] is not None and self.constraints['m'] is not None:
            if self.throw_error and not isclose(abs(self.constraints['m'] / self.constraints['s']), d, abs_tol=.01):
                raise ValueError("Provided mean and standard deviation will not produce desired inference "
                                 "statistics. Pick a different mean and/or standard deviation or remove "
                                 "them altogether.")

        if m is None and s is None:
            s = uniform(self.bounds['s'][0], self.bounds['s'][1])
            sign = -1 if uniform(self.bounds['m']
                                 [0], self.bounds['m'][1]) < 0 else 1
            m = sign * d * s
        elif m is None and self.bounds['s'][0] < s < self.bounds['s'][1]:
            self.bounds['m'][0] = d * s
            self.bounds['m'][1] = d * s
            m = d * s
        elif s is None and self.bounds['m'][0] < m < self.bounds['m'][1]:
            self.bounds['s'][0] = abs(m / d)
            self.bounds['s'][1] = abs(m / d)
            s = abs(mean / d)

        return m, s

    def _corr_const(self, d):
        low_m_s = min(abs(self.bounds['m'][0] / d),
                      abs(self.bounds['m'][1] / d))
        high_m_s = max(abs(self.bounds['m'][0] / d),
                       abs(self.bounds['m'][1] / d))

        if self.bounds['m'][0] < 0 < self.bounds['m'][1]:
            low_m_s = 0

        if low_m_s > self.bounds['s'][1] or high_m_s < self.bounds['s'][0]:
            if self.throw_error:
                raise ValueError("Provided bounds constrain the problem too much. Either loosen the "
                                 "bounds or pick different inference statistics.")
            else:
                self.bounds['s'][0] = low_m_s
                self.bounds['s'][1] = high_m_s
                return False

        if low_m_s > self.bounds['s'][0]:
            self.bounds['s'][0] = low_m_s

        if high_m_s < self.bounds['s'][1]:
            self.bounds['s'][1] = high_m_s

        sign_low_m = -1 if self.bounds['m'][0] < 0 else 1
        sign_high_m = -1 if self.bounds['m'][1] < 0 else 1

        self.bounds['m'][0] = min(max(self.bounds['m'][0], sign_low_m *
                                      self.bounds['s'][0] * d), sign_high_m * self.bounds['s'][1] * d)
        self.bounds['m'][1] = max(min(self.bounds['m'][1], sign_high_m *
                                      self.bounds['s'][1] * d), sign_low_m * self.bounds['s'][0] * d)
        return True
