from math import sqrt, isclose
from random import uniform

from scipy import stats

from cheno.util import _gen_param


class StudentD(object):

    def __init__(self, low_n=5, high_n=100, low_s=.1, high_s=5, low_m=-5, high_m=5, throw_error=True):
        self.low_n = low_n
        self.high_n = high_n
        self.low_s = low_s
        self.high_s = high_s
        self.low_m = low_m
        self.high_m = high_m
        self.throw_error = throw_error

    def solve(self, p, d, n=None, mean=None, s=None, tails=1):
        if tails != 1 and tails != 2:
            if self.throw_error:
                raise ValueError(
                    "Invalid number of tails. NHST accept either one tail or two tails.")
            else:
                tails = 2

        n = self._find_sample_sz(p, d, n, tails)
        t = abs(stats.t.ppf(1 - (p / tails), n - 1))
        mean, s = self._find_mean_std(abs(t / sqrt(n)), n, mean, s)

        return mean, s, n, _gen_param([mean], [s], [n]),

    def _find_sample_sz(self, p, d, n, tails):
        n_prime = None
        min_diff = float("inf")
        for i in range(self.low_n, self.high_n + 1):
            t_from_p = abs(stats.t.ppf(1 - (p / tails), i - 1))
            t_from_d = abs(d * sqrt(i))
            cur_diff = abs(t_from_p - t_from_d)
            if cur_diff < min_diff:
                min_diff = cur_diff
                n_prime = i

        if n is not None and n != n_prime:
            if self.throw_error:
                raise ValueError("Provided sample size will not produce desired inference statistics. "
                                 "Pick a different sample size or remove it altogether.")

        return n_prime

    def _find_mean_std(self, d, n, mean=None, s=None):
        self._corr_const(d)
        if s is not None and mean is not None and not isclose(abs(mean / s), d, abs_tol=.01):
            if self.throw_error:
                raise ValueError("Provided mean and standard deviation will not produce desired inference "
                                 "statistics. Pick a different mean and/or standard deviation or remove "
                                 "them altogether.")
            else:
                mean = None
                s = None

        if mean is None and s is None:
            s = uniform(self.low_s, self.high_s)
            sign = -1 if uniform(self.low_m, self.high_m) < 0 else 1
            mean = sign * d * s
        elif mean is None and self.low_s < s < self.high_s:
            self.low_m = d * s
            self.high_m = d * s
            mean = d * s
        elif s is None and self.low_m < mean < self.high_m:
            self.low_s = abs(mean / d)
            self.high_s = abs(mean / d)
            s = abs(mean / d)

        return mean, s

    def _corr_const(self, d):
        low_m_s = min(abs(self.low_m / d), abs(self.high_m / d))
        high_m_s = max(abs(self.low_m / d), abs(self.high_m / d))

        if self.low_m < 0 < self.high_m:
            low_m_s = 0

        if low_m_s > self.high_s or high_m_s < self.low_s:
            if self.throw_error:
                raise ValueError("Provided bounds constrain the problem too much. Either loosen the "
                                 "bounds or pick different inference statistics.")
            else:
                self.low_s = low_m_s
                self.high_s = high_m_s
                return False

        if low_m_s > self.low_s:
            self.low_s = low_m_s

        if high_m_s < self.high_s:
            self.high_s = high_m_s

        sign_low_m = -1 if self.low_m < 0 else 1
        sign_high_m = -1 if self.high_m < 0 else 1

        self.low_m = min(max(self.low_m, sign_low_m *
                             self.low_s * d), sign_high_m * self.high_s * d)
        self.high_m = max(min(self.high_m, sign_high_m *
                              self.high_s * d), sign_low_m * self.low_s * d)
        return True

    def get_const(self):
        return {"mean": [self.low_m, self.high_m], "std": [self.low_s, self.high_s], "size": [self.low_n, self.high_n]}


if __name__ == "__main__":
    # Example
    p = 5.999289792924703e-15
    es = 0.9529436452481665
    tails = 2

    # Commands
    s = StudentD()
    m, s, n, data = s.solve(p, es, tails=tails)

    # Results
    _, p_prime = stats.ttest_1samp(data[0], 0)
    print(p-p_prime)
