from itertools import product
from math import isclose, sqrt
from random import uniform

from scipy import stats

from cheno.util import _gen_param, count_non_null, count_none


class StudentI(object):

    def __init__(self, n1=[5, 100], n2=[5, 100], s1=[.1, 2], s2=[.1, 2], m1=[0, 7], m2=[0, 7], throw_error=True):
        self.n1 = n1
        self.n2 = n2
        self.s1 = s1
        self.s2 = s2
        self.m1 = m1
        self.m2 = m2
        self.throw_error = throw_error

    def solve(self, p, d, ns=None, means=None, sds=None, tails=2, eq_n=False):
        if tails != 1 and tails != 2:
            if self.throw_error:
                raise ValueError(
                    "Invalid number of tails. NHST accept either one tail or two tails.")
            else:
                tails = 2

        ns, means, sds = self._init_parameters(ns, means, sds)

        ns = self._find_sample_szs(p, d, ns, tails, eq_n)

        t = stats.t.ppf(1 - (p / tails), sum(ns) - 2)
        corr_d = abs(t * sqrt(1 / ns[0] + 1 / ns[1]))
        means, sds = self._find_means_stds(corr_d, ns, means, sds)

        return means, sds, ns, _gen_param(means, sds, ns)

    def _init_parameters(self, ns, means, sds):
        means = [None, None] if means is None else means
        sds = [None, None] if sds is None else sds
        ns = [None, None] if ns is None else ns

        # General check on the size of the constraints
        if len(ns) != 2:
            if self.throw_error:
                raise ValueError(
                    "Provided number of sample size doesn't match number of groups.")
            else:
                ns = [None, None]
        if len(means) != 2:
            if self.throw_error:
                raise ValueError(
                    "Provided number of means doesn't match number of groups")
            else:
                means = [None, None]
        if len(sds) != 2:
            if self.throw_error:
                raise ValueError(
                    "Provided number of standard deviations doesn't match number of groups")
            else:
                sds = [None, None]

        return ns, means, sds

    def _find_sample_szs(self, p, d, ns, tails, eq_n):
        # Both user sample sizes provided
        ns_comb = [ns]

        # Single user sample size provided:
        if count_none(ns) == 1:
            idx_prov_n, prov_n = (
                [(i, x) for i, x in enumerate(ns) if x is not None])[0]
            ns_comb = [[prov_n, prov_n]]
            if not eq_n:
                ns_comb = product([prov_n], range(self.n2[0], self.n2[1]))
                if idx_prov_n == 1:
                    ns_comb = product(range(self.n1[0], self.n1[1]), [prov_n])

        # No user sample sizes provided
        if count_none(ns) == 2:
            ns_comb = [[x, x] for x in range(self.n1[0], self.n1[1] + 1)]
            if not eq_n:
                ns_comb = product(
                    range(self.n1[0], self.n1[1] + 1), range(self.n2[0], self.n2[1]))

        # Minimize distance between t-value produced from inverse t-dist using p-value,
        # and t-value produced from the Cohen's d effect size eq.
        min_diff = float("inf")
        ns_prime = None
        for i in ns_comb:
            t_from_p = stats.t.ppf(1 - (p / tails), sum(i) - 2)
            t_from_d = d / sqrt(1 / i[0] + 1 / i[1])
            cur_diff = abs(abs(t_from_p) - abs(t_from_d))
            if cur_diff < min_diff:
                min_diff = cur_diff
                ns_prime = i

        if (count_none(ns) == 1 or count_none(ns) == 1) and not isclose(min_diff, 0, rel_tol=1):
            if self.throw_error:
                raise ValueError("Provided sample size/s will not produce desired inference statistics. "
                                 "Pick different sample size/s or remove it/them altogether.")

        return list(ns_prime)

    def _find_means_stds(self, d, ns, means, sds, f_alpha=0.05, tails=2):
        # F-test for homogeneity of variances
        low_f = stats.f.ppf(f_alpha / tails, ns[0] - 1, ns[1] - 1)
        high_f = stats.f.ppf(1 - f_alpha / tails, ns[0] - 1, ns[1] - 1)

        # Both means and both standard deviations provided
        if count_non_null(means) == 2 and count_non_null(sds) == 2:
            if not low_f < (sds[0] / sds[1]) ** 2 < high_f:
                if self.throw_error:
                    raise ValueError("Provided standard deviations are not equal. Either try Welch's t-test "
                                     "or correct your standard deviations.")
            sd_p = sqrt(((ns[0] - 1) * sds[0] ** 2 + (ns[1] - 1)
                         * sds[1] ** 2) / (ns[0] + ns[1] - 2))
            mean_diff = means[0] - means[1]
            if not isclose(d, mean_diff / sd_p, abs_tol=0.001):
                if self.throw_error:
                    raise ValueError("Provided means and standard deviations will not produce desired inference "
                                     "statistics. Pick different means and/or standard deviations or remove "
                                     "them altogether.")
                else:
                    means = [None, None]
                    sds = [None, None]

        #  Both means provided
        if count_non_null(means) == 2:
            mean_diff = means[0] - means[1]
            sd_p = mean_diff / d
            if count_non_null(sds) == 1:
                prov_sd_idx, prov_sd = (
                    [(i, x) for i, x in enumerate(sds) if x is not None])[0]
                sds[1 - prov_sd_idx] = sqrt(
                    (sd_p * (ns[0] + ns[1] - 2) - (ns[1 - prov_sd_idx] - 1) * prov_sd ** 2) / (ns[prov_sd_idx] - 1))
            else:
                sds[0] = (sd_p / 2) * uniform(1, 1.05)
                sds[1] = sqrt((sd_p * (ns[0] + ns[1] - 2) -
                               (ns[0] - 1) * sds[0] ** 2) / (ns[1] - 1))
            return means, sds

        # Both standard deviations are provided
        if count_none(sds) == 0:
            if not low_f < (sds[0] / sds[1]) ** 2 < high_f:
                if self.throw_error:
                    raise ValueError("Provided standard deviations are not equal. Either try Welch's t-test "
                                     "or correct your standard deviations.")
                else:
                    sds = [None, None]

        # Single standard deviation provided
        if count_none(sds) == 1:
            prov_sd_idx, prov_sd = (
                [(i, x) for i, x in enumerate(sds) if x is not None])[0]
            if prov_sd_idx == 0:
                self.s2 = [max(sds[0] / sqrt(high_f), self.s2[0]),
                           min(sds[0] / sqrt(low_f), self.s2[1])]
            else:
                self.s1 = [max(sqrt(low_f) * sds[1], self.s1[0]),
                           min(sqrt(high_f) * sds[1], self.s1[1])]
            sds[1 - prov_sd_idx] = uniform(self.s1[0], self.s1[1]) if prov_sd_idx == 1 else uniform(self.s2[0],
                                                                                                    self.s2[1])
        # No standard deviation provided
        else:
            sds[0] = uniform(self.s1[0], self.s1[1])
            self.s2 = [max(sds[0] / sqrt(high_f), self.s2[0]),
                       min(sds[0] / sqrt(low_f), self.s2[1])]
            sds[1] = uniform(self.s2[0], self.s2[1])

        sd_p = sqrt(((ns[0] - 1) * sds[0] ** 2 + (ns[1] - 1)
                     * sds[1] ** 2) / (ns[0] + ns[1] - 2))
        mean_diff = d * sd_p

        if count_none(means) == 1:
            prov_mean_idx, prov_mean = (
                [(i, x) for i, x in enumerate(means) if x is not None])[0]
            means[1 - prov_mean_idx] = mean_diff + \
                prov_mean if prov_mean_idx == 1 else prov_mean - mean_diff

        if count_none(means) == 2:
            means[0] = uniform(self.m1[0], self.m2[0])
            means[1] = means[0] - mean_diff

        return means, sds

    def get_const(self):
        return {"means": [self.m1, self.m2], "stds": [self.s1, self.s2], "sizes": [self.n1, self.n2]}


if __name__ == "__main__":
    # Example
    p = .107
    d = -0.6862653762697537

    # Commands
    s = StudentI()
    means, stds, ns, data = s.solve(
        p, d, means=[0.5, None], sds=[0.3, None])

    # Results
    _, p_prime = stats.ttest_ind(data[0], data[1])
    print(p_prime)
    print(s.get_const())
