from math import sqrt

import numpy.core.numeric as numeric
from numpy import mean, std
from numpy.random import uniform, normal
from scipy.stats import ttest_1samp, ttest_ind, f_oneway, mannwhitneyu, wilcoxon, kruskal, friedmanchisquare, norm


def _eft_sz_studentd(t, n):
    return t / sqrt(n)


def _eft_sz_studenti_eq_n(t, ns):
    return t * sqrt(1 / ns[0] + 1 / ns[1])


def _eft_sz_welch(x, y):
    m1, s1, n1 = mean(x), std(x, ddof=1), len(x)
    m2, s2, n2 = mean(y), std(y, ddof=1), len(y)
    dof = (n1 + n2 - 2)
    d = (m1 - m2) / sqrt(((n1 - 1) * s1 ** 2 + (n2 - 1) * s2 ** 2) / dof)
    return d


def _eft_sz_anova(*args):
    ns = []
    means = []
    sds = []
    for x in args:
        ns.append(len(x))
        means.append(numeric.mean(x))
        sds.append(numeric.std(x, ddof=1))
    g_mean = sum([m * n for m, n in zip(means, ns)]) / sum(ns)
    df_m = len(args) - 1
    df_r = sum(ns) - len(args)
    ss_m = sum([n * (m - g_mean) ** 2 for m, n in zip(means, ns)])
    ss_r = sum((n - 1) * s ** 2 for s, n in zip(sds, ns))
    return ss_m / (ss_m + ss_r)


def _eft_sz_mann_whitney(p, n, tails=1):
    z = norm.ppf(p / tails)
    return z / sqrt(n)


def _eft_sz_wilcoxon(p, n, tails=1):
    z = norm.ppf(p / tails)
    return z / sqrt(n)


def _eft_sz_kruskall(h, n, k):
    return (h - k + 1) / (n - k)


def _eft_sz_friedman(f, n, k):
    return f / (n * (k - 1))


def gen_student_dep(mean=0.042, s=1.77, n=100):
    x = normal(mean, s, n)
    t, p = ttest_1samp(x, 0)
    d = _eft_sz_studentd(t, n)
    with open('data/student_d.txt', 'a') as f:


def gen_student_ind_eq_n(mean1=0.042, mean2=0.043, s=1.77, n=100):
    x = normal(mean1, s, n)
    y = normal(mean2, s, n)
    t, p = ttest_ind(x, y)
    d = _eft_sz_studenti_eq_n(t, n)
    with open('data/student_i_eq_n.txt', 'a') as f:
        f.write(str(p) + " " + str(d) + "\n")
        f.write(str(numeric.mean(x)) + " " + str(numeric.std(x, ddof=1)) + "\n")


def gen_student_ind_uneq_n(mean1=0.042, mean2=0.043, s=1.77, n1=100, n2=100):
    x = normal(mean1, s, n1)
    y = normal(mean2, s, n2)
    t, p = ttest_ind(x, y)
    d = _eft_sz_welch(x, y)
    with open('data/student_i_uneq_n.txt', 'a') as f:
        f.write(str(p) + " " + str(d) + "\n")
        f.write(str(numeric.mean(x)) + " " + str(numeric.std(x, ddof=1)) + "\n")


def gen_welch(mean1=0.042, mean2=0.043, s1=1.77, s2=5.6, n=100):
    x = normal(mean1, s1, n)
    y = normal(mean2, s2, n)
    t, p = ttest_ind(x, y, equal_var=False)
    d = _eft_sz_welch(x, y)
    with open('data/welch.txt', 'a') as f:
        f.write(str(p) + "\n " + str(d) + "\n")


def gen_anova(k=2, mean=0.0043, s=1.77, n=100):
    samples = []
    for _ in range(k):
        samples.append(normal(mean, s, n))
    f, p = f_oneway(*samples)
    eta2 = _eft_sz_anova(*samples)
    with open(f'data/anova_eq_n_{k}_{n}.txt', 'a') as f:
        f.write(str(p) + "\n")
        f.write(str(eta2) + "\n")


def gen_mann_whitney(lb=-5, ub=5, n=100, shape=uniform):
    x = uniform(lb, ub, n)
    y = uniform(lb, ub, n)
    _, p = mannwhitneyu(x, y)
    r = _eft_sz_mann_whitney(p, 2 * n, 2)
    with open('data/mann_whitney.txt', 'a') as f:
        f.write(str(p) + "\n")
        f.write(str(r) + "\n")


def gen_wilcoxon(lb=-5, ub=5, n=100, shape=uniform):
    x = uniform(lb, ub, n)
    y = uniform(lb, ub, n)
    _, p = wilcoxon(x, y)
    r = _eft_sz_wilcoxon(p, 2 * n, 2)
    with open('data/wilcoxon.txt', 'a') as f:
        f.write(str(p) + "\n")
        f.write(str(r) + "\n")


def gen_kruskal_wallis(lb=-5, ub=5, n=100, k=2, shape=uniform):
    samples = []
    for _ in range(k):
        samples.append(uniform(lb, ub, n))
    h, p = kruskal(*samples)
    eta2 = _eft_sz_kruskall(h, k * n, k)
    with open(f'data/kruskal_wallis_{k}_{n}.txt', 'a') as f:
        f.write(str(p) + "\n")
        f.write(str(eta2) + "\n")


def gen_friedman(lb=-5, ub=5, n=100, k=3, shape=uniform):
    samples = []
    for _ in range(k):
        samples.append(uniform(lb, ub, n))
    f, p = friedmanchisquare(*samples)
    w = _eft_sz_friedman(f, n, k)
    with open(f'data/friedman_{k}_{n}.txt', 'a') as f:
        f.write(str(p) + "\n")
        f.write(str(w) + "\n")


if __name__ == "__main__":
    for i in range(10000):
        # gen_student_dep(mean=(i / 100000))
        gen_student_ind_eq_n(mean1=0.042, mean2=(0.042 + (i / 100000)))
        gen_student_ind_uneq_n(mean1=0.042, mean2=(0.042 + (i / 100000)), n1=100, n2=100 + (i % 100))
        gen_welch(mean2=0.042 + (i % 1000) / 100000, s2=5.6+(i/10000))
        # gen_anova(mean=(0.042 + (i / 100000)), k=4)
        # gen_anova(mean=(0.042 + (i / 100000)), k=7)
        # gen_mann_whitney()
        # gen_wilcoxon()
        # gen_kruskal_wallis(k=4, n=100)
        # gen_kruskal_wallis(k=7, n=100)
        # gen_friedman(k=4, n=100)
        # gen_friedman(k=7, n=100)
