from cheno import StudentD, StudentI, Welch, ANOVA, MannWhitney, Wilcoxon, KruskalWallis, Friedman
import cheno.tests.util as util

from time import time
import logging
import traceback
import logging
import datetime
import os
from statistics import mean

import numpy as np
from scipy import stats

# initialize the log settings
filename = os.path.splitext(os.path.basename(__file__))[0]
logging.basicConfig(filename=f'log/{filename}_{datetime.datetime.today().strftime("%y%m%d_%H%M")}.log',
                    level=logging.INFO)


def log_results(test, times, rel_tol_stats, rel_tol_data, error_cases):
    logging.info(f'\n------------------------------------------------')
    logging.info(f'RESULTS FOR: {test}')
    logging.info(f'Measure  , Min, Mean, Max')
    logging.info(f'Time     , {min(times)}, {np.nanmean(times)}, {max(times)}')
    logging.info(f'AccStats , {min(rel_tol_stats)}, {np.nanmean(rel_tol_stats)}, {max(rel_tol_stats)}')
    logging.info(f'AccData  , {min(rel_tol_data)}, {np.nanmean(rel_tol_data)}, {max(rel_tol_data)}')
    logging.info(f'------------------------------------------------')
    logging.info(f'Error Cases: {len(error_cases)}')
    for error in error_cases:
        logging.info(error)
    logging.info(f'------------------------------------------------\n')


def generate_data_studentD(epochs):
    lines = []
    for _ in range(epochs):
        data = util._sample_data_for_test('StudentD')
        t, p = stats.ttest_1samp(data[0], 0)
        m = np.mean(data[0])
        s = np.std(data[0], ddof=1)
        n = data[0].size
        ef = util._eft_sz_studentd(t, n)
        dstats = [m, s, n]
        lines.append((p, ef, dstats))

    with open(f'data/StudentD.txt', 'w') as f:
        for line in lines:
            dstats = ", ".join([str(x) for x in line[2]])
            f.write(f'{line[0]}, {line[1]}, 2, {dstats}\n')


def evaluate_studentD(epochs):
    logging.info(f'Started testing: {StudentD}')
    with open(f"data/StudentD.txt", "r") as f:
        rel_tol_stats = []
        rel_tol_data = []
        times = []
        error_cases = []
        for i, line in enumerate(f):
            util.progress_bar(i, epochs)
            # Get data
            stats = line.strip().split(', ')
            p, ef, tails = float(stats[0]), float(stats[1]), int(stats[2])
            # Run test
            try:
                measurements = run_studentD(p, ef, tails)
                # Enter measured values
                times.append(measurements[0])
                rel_tol_stats.append(measurements[1])
                rel_tol_data.append(measurements[2])
            except Exception as e:
                error_cases.append(line)
                logging.error("Error occurred" + str(e))
                logging.error(f'INPUT:\np = {p}\nes = {ef}\ntails= {tails}\n')
                logging.error(traceback.format_exc())
        log_results("StudentD", times, rel_tol_stats, rel_tol_data, error_cases)


def run_studentD(p, d, tails):
    student = StudentD()
    start = time()
    m, s, n, data = student.solve(p, d, tails=tails)
    end = time()
    _, p_from_stats = util._d_ttest_from_stats(m, s, n)
    _, p_from_data = stats.ttest_1samp(data[0], 0)
    return (end - start), abs((p - p_from_stats) / p), abs((p - p_from_data) / p)


def generate_data_studentI(epochs):
    lines = []
    for _ in range(epochs):
        data = util._sample_data_for_test('StudentI')
        t, p = stats.ttest_ind(data[0], data[1])

        ms = [np.mean(data[0]), np.mean(data[1])]
        ss = [np.std(data[0], ddof=1), np.std(data[1], ddof=1)]
        ns = [data[0].size, data[1].size]
        ef = util._eft_sz_studenti_eq_n(t, ns)
        dstats = [*ms, *ss, *ns]
        lines.append((p, ef, dstats))

    with open(f'data/StudentI.txt', 'w') as f:
        for line in lines:
            dstats = ", ".join([str(x) for x in line[2]])
            f.write(f'{line[0]}, {line[1]}, 2, {dstats}\n')


def evaluate_studentI(epochs):
    logging.info(f'Started testing: StudentI')
    with open(f"data/StudentI.txt", "r") as f:
        rel_tol_stats = []
        rel_tol_data = []
        times = []
        error_cases = []
        for i, line in enumerate(f):
            util.progress_bar(i, epochs)
            # Get data
            stats = line.strip().split(', ')
            p, ef, tails = float(stats[0]), float(stats[1]), int(stats[2])
            # Run test
            try:
                measurements = run_studentI(p, ef, tails)
                # Enter measured values
                times.append(measurements[0])
                rel_tol_stats.append(measurements[1])
                rel_tol_data.append(measurements[2])
            except Exception as e:
                error_cases.append(line)
                logging.error("Error occurred" + str(e))
                logging.error(f'INPUT:\np = {p}\nes = {ef}\ntails= {tails}\n')
                logging.error(traceback.format_exc())
        log_results("StudentI", times, rel_tol_stats, rel_tol_data, error_cases)


def run_studentI(p, d, tails):
    s = StudentI()
    start = time()
    ms, ss, ns, data = s.solve(p, d, tails=tails)
    end = time()
    _, p_from_stats = stats.ttest_ind_from_stats(ms[0], ss[0], ns[0], ms[1], ss[1], ns[1])
    _, p_from_data = stats.ttest_ind(data[0], data[1])
    return (end - start), abs((p - p_from_stats) / p), abs((p - p_from_data) / p)


def generate_data_welch(epochs):
    lines = []
    for _ in range(epochs):
        data = util._sample_data_for_test('Welch')
        t, p = stats.ttest_ind(data[0], data[1])

        ms = [np.mean(data[0]), np.mean(data[1])]
        ss = [np.std(data[0], ddof=1), np.std(data[1], ddof=1)]
        ns = [data[0].size, data[1].size]
        ef = util._eft_sz_welch(*data)
        dstats = [*ms, *ss, *ns]
        lines.append((p, ef, dstats))

    with open(f'data/Welch.txt', 'w') as f:
        for line in lines:
            dstats = ", ".join([str(x) for x in line[2]])
            f.write(f'{line[0]}, {line[1]}, 2, {dstats}\n')


def evaluate_welch(epochs):
    logging.info(f'Started testing: Welch')
    with open(f"data/Welch.txt", "r") as f:
        rel_tol_stats = []
        rel_tol_data = []
        times = []
        error_cases = []
        for i, line in enumerate(f):
            util.progress_bar(i, epochs)
            # Get data
            stats = line.strip().split(', ')
            p, ef, tails = float(stats[0]), float(stats[1]), int(stats[2])
            # Run test
            try:
                measurements = run_welch(p, ef, tails)
                # Enter measured values
                times.append(measurements[0])
                rel_tol_stats.append(measurements[1])
                rel_tol_data.append(measurements[2])
            except Exception as e:
                error_cases.append(line)
                logging.error("Error occurred" + str(e))
                logging.error(f'INPUT:\np = {p}\nes = {ef}\ntails= {tails}\n')
                logging.error(traceback.format_exc())
        log_results("Welch", times, rel_tol_stats, rel_tol_data, error_cases)


def run_welch(p, d, tails):
    s = Welch()
    start = time()
    ms, ss, ns, data = s.solve(p, d, tails=tails, splits=10)
    end = time()
    _, p_from_stats = stats.ttest_ind_from_stats(ms[0], ss[0], ns[0], ms[1], ss[1], ns[1], equal_var=False)
    _, p_from_data = stats.ttest_ind(data[0], data[1], equal_var=False)
    return (end - start), abs((p - p_from_stats) / p), abs((p - p_from_data) / p)


def generate_data_anova(epochs):
    lines = []
    for _ in range(epochs):
        data = util._sample_data_for_test('ANOVA')
        t, p = stats.f_oneway(*data)

        ms = [np.mean(d) for d in data]
        ss = [np.std(d, ddof=1) for d in data]
        ns = [d.size for d in data]
        ef = util._eft_sz_anova(*data)
        dstats = [*ms, *ss, *ns]
        lines.append((p, ef, dstats))

    with open(f'data/ANOVA.txt', 'w') as f:
        for line in lines:
            dstats = ", ".join([str(x) for x in line[2]])
            f.write(f'{line[0]}, {line[1]}, 2, {dstats}\n')


def evaluate_anova(epochs):
    logging.info(f'Started testing: ANOVA')
    with open(f"data/ANOVA.txt", "r") as f:
        rel_tol_stats = []
        rel_tol_data = []
        times = []
        error_cases = []
        for i, line in enumerate(f):
            util.progress_bar(i, epochs)
            # Get data
            stats = line.strip().split(', ')
            p, ef, tails = float(stats[0]), float(stats[1]), int(stats[2])
            # Run test
            try:
                measurements = run_anova(p, ef)
                # Enter measured values
                times.append(measurements[0])
                rel_tol_stats.append(measurements[1])
                rel_tol_data.append(measurements[2])
            except Exception as e:
                error_cases.append(line)
                logging.error("Error occurred: " + str(e))
                logging.error(f'INPUT:\np = {p}\nes = {ef}\ntails= {tails}\n')
                logging.error(traceback.format_exc())
        log_results("ANOVA", times, rel_tol_stats, rel_tol_data, error_cases)


def run_anova(p, eta2):
    s = ANOVA(k=7)
    start = time()
    ms, ss, ns, data = s.solve(p, eta2)
    end = time()
    _, p_from_stats = stats.f_oneway(*data)
    _, p_from_data = stats.f_oneway(*data)
    return (end - start), abs((p - p_from_stats) / p), abs((p - p_from_data) / p)


def generate_data_mannwhitney(epochs):
    lines = []
    for _ in range(epochs):
        data = util._sample_data_for_test('MannWhitney')
        t, p = stats.mannwhitneyu(*data)

        ms = [np.mean(d) for d in data]
        ss = [np.std(d, ddof=1) for d in data]
        ns = [d.size for d in data]
        ef = util._eft_sz_mann_whitney(p, sum(ns), 2)
        dstats = [*ms, *ss, *ns]
        lines.append((p, ef, dstats))

    with open(f'data/MannWhitney.txt', 'w') as f:
        for line in lines:
            dstats = ", ".join([str(x) for x in line[2]])
            f.write(f'{line[0]}, {line[1]}, 2, {dstats}\n')


def evaluate_mannwhitney(epochs):
    logging.info(f'Started testing: MannWhitney')
    with open(f"data/MannWhitney.txt", "r") as f:
        rel_tol_stats = []
        rel_tol_data = []
        times = []
        error_cases = []
        for i, line in enumerate(f):
            util.progress_bar(i, epochs)
            # Get data
            stats = line.strip().split(', ')
            p, ef, tails = float(stats[0]), float(stats[1]), int(stats[2])
            # Run test
            try:
                measurements = run_mannwhitney(p, ef, tails)
                # Enter measured values
                times.append(measurements[0])
                rel_tol_stats.append(measurements[1])
                rel_tol_data.append(measurements[2])
            except Exception as e:
                error_cases.append(line)
                logging.error("Error occurred: " + str(e))
                logging.error(f'INPUT:\np = {p}\nes = {ef}\ntails= {tails}\n')
                logging.error(traceback.format_exc())
        log_results("MannWhitney", times, rel_tol_stats, rel_tol_data, error_cases)


def run_mannwhitney(p, r, tails):
    s = MannWhitney()
    start = time()
    ranks, data = s.solve(p, r, tails=tails)
    end = time()
    p_from_stats = util._mann_whitney_from_stats(ranks)
    _, p_from_data = stats.mannwhitneyu(*data, alternative="two-sided")
    return (end - start), abs((p - p_from_stats) / p), abs((p - p_from_data) / p)


def generate_data_wilcoxon(epochs):
    lines = []
    for _ in range(epochs):
        data = util._sample_data_for_test('Wilcoxon')
        t, p = stats.wilcoxon(*data, zero_method='wilcox')

        ms = [np.mean(d) for d in data]
        ss = [np.std(d, ddof=1) for d in data]
        ns = [d.size for d in data]
        ef = util._eft_sz_wilcoxon(p, sum(ns), 2)
        dstats = [*ms, *ss, *ns]
        lines.append((p, ef, dstats))

    with open(f'data/Wilcoxon.txt', 'w') as f:
        for line in lines:
            dstats = ", ".join([str(x) for x in line[2]])
            f.write(f'{line[0]}, {line[1]}, 2, {dstats}\n')


def evaluate_wilcoxon(epochs):
    logging.info(f'Started testing: Wilcoxon')
    with open(f"data/Wilcoxon.txt", "r") as f:
        rel_tol_stats = []
        rel_tol_data = []
        times = []
        error_cases = []
        for i, line in enumerate(f):
            util.progress_bar(i, epochs)
            # Get data
            stats = line.strip().split(', ')
            p, ef, tails = float(stats[0]), float(stats[1]), int(stats[2])
            # Run test
            try:
                measurements = run_wilcoxon(p, ef, tails)
                # Enter measured values
                times.append(measurements[0])
                rel_tol_stats.append(measurements[1])
                rel_tol_data.append(measurements[2])
            except Exception as e:
                error_cases.append(line)
                logging.error("Error occurred: " + str(e))
                logging.error(f'INPUT:\np = {p}\nes = {ef}\ntails= {tails}\n')
                logging.error(traceback.format_exc())
        log_results("Wilcoxon", times, rel_tol_stats, rel_tol_data, error_cases)


def run_wilcoxon(p, r, tails):
    s = Wilcoxon()
    start = time()
    ranks, data = s.solve(p, r, tails=tails)
    end = time()
    p_from_stats = util._wilcoxon_from_stats(ranks)
    _, p_from_data = stats.wilcoxon(*data, zero_method='wilcox')
    return (end - start), abs((p - p_from_stats) / p), abs((p - p_from_data) / p)


def generate_data_kruskalwallis(epochs):
    lines = []
    for _ in range(epochs):
        data = util._sample_data_for_test('KruskalWallis')
        t, p = stats.kruskal(*data)

        ms = [np.mean(d) for d in data]
        ss = [np.std(d, ddof=1) for d in data]
        ns = [d.size for d in data]
        ef = util._eft_sz_kruskall(t, sum(ns), 7)
        dstats = [*ms, *ss, *ns]
        lines.append((p, ef, dstats))

    with open(f'data/KruskalWallis.txt', 'w') as f:
        for line in lines:
            dstats = ", ".join([str(x) for x in line[2]])
            f.write(f'{line[0]}, {line[1]}, 2, {dstats}\n')


def evaluate_kruskalwallis(epochs):
    logging.info(f'Started testing: KruskalWallis')
    with open(f"data/KruskalWallis.txt", "r") as f:
        rel_tol_stats = []
        rel_tol_data = []
        times = []
        error_cases = []
        for i, line in enumerate(f):
            util.progress_bar(i, epochs)
            # Get data
            stats = line.strip().split(', ')
            p, ef, tails = float(stats[0]), float(stats[1]), int(stats[2])
            # Run test
            try:
                measurements = run_kruskalwallis(p, ef, tails)
                # Enter measured values
                times.append(measurements[0])
                rel_tol_stats.append(measurements[1])
                rel_tol_data.append(measurements[2])
            except Exception as e:
                error_cases.append(line)
                logging.error("Error occurred: " + str(e))
                logging.error(f'INPUT:\np = {p}\nes = {ef}\ntails= {tails}\n')
                logging.error(traceback.format_exc())
        log_results("KruskalWallis", times, rel_tol_stats, rel_tol_data, error_cases)


def run_kruskalwallis(p, r, tails):
    s = KruskalWallis([None for _ in range(7)])
    start = time()
    ranks, data = s.solve(p, r, epoch=1000)
    end = time()
    p_from_stats = util._kruskal_wallis_from_stats(ranks)
    _, p_from_data = stats.kruskal(*data)
    return (end - start), abs((p - p_from_stats) / p), abs((p - p_from_data) / p)


def generate_data_friedman(epochs):
    lines = []
    for _ in range(epochs):
        data = util._sample_data_for_test('Friedman')
        t, p = stats.friedmanchisquare(*data)

        ms = [np.mean(d) for d in data]
        ss = [np.std(d, ddof=1) for d in data]
        ns = [d.size for d in data]
        ef = util._eft_sz_friedman(t, sum(ns), 7)
        dstats = [*ms, *ss, *ns]
        lines.append((p, ef, dstats))

    with open(f'data/Friedman.txt', 'w') as f:
        for line in lines:
            dstats = ", ".join([str(x) for x in line[2]])
            f.write(f'{line[0]}, {line[1]}, 2, {dstats}\n')


def evaluate_friedman(epochs):
    logging.info(f'Started testing: Friedman')
    with open(f"data/Friedman.txt", "r") as f:
        rel_tol_stats = []
        rel_tol_data = []
        times = []
        error_cases = []
        for i, line in enumerate(f):
            util.progress_bar(i, epochs)
            # Get data
            stats = line.strip().split(', ')
            p, ef, tails = float(stats[0]), float(stats[1]), int(stats[2])
            # Run test
            try:
                measurements = run_friedman(p, ef, tails)
                # Enter measured values
                times.append(measurements[0])
                rel_tol_stats.append(measurements[1])
                rel_tol_data.append(measurements[2])
            except Exception as e:
                error_cases.append(line)
                logging.error("Error occurred: " + str(e))
                logging.error(f'INPUT:\np = {p}\nes = {ef}\ntails= {tails}\n')
                logging.error(traceback.format_exc())
        log_results("Friedman", times, rel_tol_stats, rel_tol_data, error_cases)


def run_friedman(p, r, tails):
    s = Friedman()
    start = time()
    ranks, data = s.solve(p, r, 7, epoch=1000)
    end = time()
    p_from_stats = util._friedman_from_stats(ranks)
    _, p_from_data = stats.friedmanchisquare(*data)
    return (end - start), abs((p - p_from_stats) / p), abs((p - p_from_data) / p)


if __name__ == "__main__":
    # StudentD
    # generate_data_studentD(1000)
    # evaluate_studentD(1000)

    # StudentI
    # generate_data_studentI(1000)
    # evaluate_studentI(1000)

    # Welch
    generate_data_welch(1000)
    evaluate_welch(1000)

    # ANOVA
    # generate_data_anova(1000)
    # evaluate_anova(1000)

    # Mann-Whitney
    # generate_data_mannwhitney(1000)
    # evaluate_mannwhitney(1000)

    # Wilcoxon
    # generate_data_wilcoxon(1000)
    # evaluate_wilcoxon(1000)

    # Kruskal-Wallis
    # generate_data_kruskalwallis(1000)
    # evaluate_kruskalwallis(1000)

    # Friedman
    # generate_data_friedman(1000)
    # evaluate_friedman(1000)
