import cheno
import datagen

from math import sqrt
from time import time
from scipy import stats 
from numpy import mean, random, std
import logging
import datetime
import traceback

PARAM = ['StudentD', 'StudentI', 'ANOVA']
NONPARAM = ['MannWhitney', 'Wilcoxon', 'KruskalWallis', 'Friedman']


def _d_ttest_from_stats(mean, s, n):
    t = mean / (s / sqrt(n))
    p = stats.t.sf(abs(t), n - 1) * 2
    return t, p


def _mann_whitney_from_stats(ranks):
    s1, s2 = sum(ranks[0]), sum(ranks[1])
    w = s1 if s1 < s2 else s2
    n1, n2 = len(ranks[0]), len(ranks[1])
    w_mean = n1 * (n1 + n2 + 1) / 2
    se_w = sqrt(n1 * n2 * (n1 + n2 + 1) / 12)
    z = (w - w_mean) / se_w
    return stats.norm.cdf(z) * 2


def _wilcoxon_from_stats(ranks):
    s1, s2 = sum(ranks[0]), sum(ranks[1])
    w = s1 if s1 < s2 else s2
    n1, n2 = len(ranks[0]), len(ranks[1])
    w_mean = min(n1, n2) * (n1 + n2 + 1) / 2
    se_w = sqrt(n1 * n2 * (n1 + n2 + 1) / 12)
    z = (w - w_mean) / se_w
    return stats.norm.cdf(z) * 2


def _kruskal_wallis_from_stats(ranks):
    s = [(sum(x), len(x)) for x in ranks]
    n = sum([x[1] for x in s])
    h = (12 / (n * (n + 1))) * sum([x[0] ** 2 / x[1] for x in s]) - 3 * (n + 1)
    return stats.chi2.sf(h, 3)


def _friedman_from_stats(ranks):
    s = [(sum(x), len(x)) for x in ranks.T]
    n = sum([x[1] for x in s]) / k
    h = (12 / (n * k * (k + 1))) * sum([x[0] ** 2 for x in s]) - 3 * n * (k + 1)
    return stats.chi2.sf(h, k - 1)    


class Suite(object):
    rel_tol_stats = []
    rel_tol_data = []
    times = []

    def __init__(self, test, path):
        # initialize the log settings
        logging.basicConfig(filename=f'cheno/tests/log/{test}_{datetime.datetime.today().strftime("%y%m%d_%H%M")}.log', level=logging.INFO)        
        self.test = test
        self.path = path

    def run_suite(self):
        logging.info(f'Started testing: {self.test}')
        with open(f"cheno/tests/{self.path}", "r") as f:
            for i, line in enumerate(f):
                stats = line.strip().split(', ')
                self._run_test(float(stats[0]), float(stats[1]), int(stats[2]))

    def log_results(self):
        logging.info(f'RESULTS FOR: {self.test}')
        logging.info(f'Measure  , Min, Mean, Max')
        logging.info(f'AccStats , {min(self.rel_tol_stats)}, {mean(self.rel_tol_stats)}, {max(self.rel_tol_stats)}')
        logging.info(f'AccData  , {min(self.rel_tol_data)}, {mean(self.rel_tol_data)}, {max(self.rel_tol_data)}')
        logging.info(f'Time     , {min(self.times)}, {mean(self.times)}, {max(self.times)}')

    def _init_test(self, test):
        if test == 'StudentD':
            return cheno.StudentD()
        elif test == 'StudentI':
            return cheno.StudentI()
        elif test == 'Welch':
            return cheno.Welch()
        elif test == 'ANOVA':
            return cheno.ANOVA(k=7)
        elif test == 'MannWhitney':
            return cheno.MannWhitney()
        elif test == 'Wilcoxon':
            return cheno.Wilcoxon()
        elif test == 'KruskalWallis':
            return cheno.KruskalWallis()
        elif test == 'Friedman':
            return cheno.Friedman()

    def _run_test(self, p, es, tails):
        test = self._init_test(self.test)
        try:
            start = time()
            results = test.solve(p, es, tails=tails)
            end = time()

            # Log Time
            self.times.append(end - start)
            
            #Log Relative Error
            if self.test in PARAM:
                _, p_prime = Suite._sig_test_from_stats(self.test, results[0:3])
                self.rel_tol_stats.append(abs((p - p_prime) / p))
            
            _, p_prime = Suite._sig_test_from_data(self.test, results[-1])
            self.rel_tol_data.append(abs((p - p_prime) / p))
        except Exception as e:
            logging.error("Error occurred" + str(e))
            logging.error(f'Input p= {p} es= {es} tails= {tails}')
            logging.error(traceback.format_exc())

    @staticmethod
    def _sig_test_from_data(test, data):
        if test == 'StudentD':
            return stats.ttest_1samp(data[0], 0)
        elif test == 'StudentI':
            return stats.ttest_ind(*data)
        elif test == 'Welch':
            return stats.ttest_ind(*data, equal_var=False)
        elif test == 'ANOVA':
            return stats.f_oneway(*data)
        elif test == 'MannWhitney':
            return stats.mannwhitneyu(*data)
        elif test == 'Wilcoxon':
            return stats.wilcoxon(*data)
        elif test == 'KruskalWallis':
            return stats.kruskal(*data)
        elif test == 'Friedman':
            return stats.friedmanchisquare(*data)

    @staticmethod
    def _sig_test_from_stats(test, dstats):
        if test == 'StudentD':
            return _d_ttest_from_stats(*dstats)
        elif test == 'StudentI':
            return stats.ttest_ind_from_stats(dstats[0][0], dstats[1][0], dstats[2][0], dstats[0][1], dstats[1][1], dstats[2][1])
        elif test == 'Welch':
            return stats.ttest_ind_from_stats(dstats[0][0], dstats[1][0], dstats[2][0], dstats[0][1], dstats[1][1], dstats[2][1], equal_var=False)
        elif test == 'ANOVA':
            return cheno.ANOVA()
        elif test == 'MannWhitney':
            return _mann_whitney_from_stats(*dstats)
        elif test == 'Wilcoxon':
            return _wilcoxon_from_stats(*dstats)
        elif test == 'KruskalWallis':
            return _kruskal_wallis_from_stats(*dstats)
        elif test == 'Friedman':
            return _friedman_from_stats(*dstats)


class DataGen(object):
    
    @staticmethod
    def gen_datasets(epochs):
        tests = PARAM + NONPARAM
        for t in tests:
            lines = []
            for _ in range(epochs):
                data = DataGen._sample_data_for_test(t)
                lines.append(DataGen._compute_stats_from_data(t, data))
                
            with open(f'cheno/tests/data/{t}.txt', 'w') as f:
                for line in lines:
                    dstats = ", ".join([str(x) for x in line[2]])
                    f.write(f'{line[0]}, {line[1]}, 2, {dstats}\n')

    @staticmethod
    def _sample_data_for_test(test):
        if test == 'StudentD':
            t = cheno.StudentD()
            m = random.uniform(t.low_m, t.high_m)
            s = random.uniform(t.low_s, t.high_s)
            n = random.randint(t.low_n, t.high_n + 1)
            data = random.normal(m, s, n)
            return [data]
        elif test == 'StudentI':
            t = cheno.StudentI()
            ms = [random.uniform(m[0], m[1]) for m in [t.m1, t.m2]]
            ss = [random.uniform(m[0], m[1]) for m in [t.s1, t.s2]]
            ns = [random.randint(m[0], m[1] + 1) for m in [t.n1, t.n2]]
            return [random.normal(m, s, n) for m, s, n in zip(ms, ss, ns)]
        elif test == 'Welch':
            t = cheno.Welch()
            ms = [random.uniform(m[0], m[1]) for m in [t.m1, t.m2]]
            ss = [random.uniform(m[0], m[1]) for m in [t.s1, t.s2]]
            ns = [random.randint(m[0], m[1] + 1) for m in [t.n1, t.n2]]
            return [random.normal(m, s, n) for m, s, n in zip(ms, ss, ns)]
        elif test == 'ANOVA':
            t = cheno.ANOVA(k=7)
            ms = [random.uniform(m[0], m[1]) for m in t.m_bounds]
            ss = [random.uniform(m[0], m[1]) for m in t.s_bounds]
            ns = [random.randint(m[0], m[1] + 1) for m in t.n_bounds]
            return [random.normal(m, s, n) for m, s, n in zip(ms, ss, ns)]
        elif test == 'MannWhitney':
            t = cheno.MannWhitney()
            ns = [random.randint(m[0], m[1] + 1) for m in [(t.low_n, t.high_n), (t.low_n, t.high_n)]]
            return [random.sample(n) for n in ns]
        elif test == 'Wilcoxon':
            t = cheno.Wilcoxon()
            n = random.randint(t.lb, t.ub + 1)
            return [random.sample(n) for _ in range(2)]
        elif test == 'KruskalWallis':
            t = cheno.ANOVA(k=7)
            ns = [random.randint(8, m[1] + 1) for m in t.n_bounds]
            return [random.sample(n) for n in ns]
        elif test == 'Friedman':
            t = cheno.ANOVA(k=7)
            n = random.randint(8, t.n_bounds[0][1] + 1)
            return [random.sample(n) for _ in range(7)]

    @staticmethod
    def _compute_stats_from_data(test, data):
        es = None
        dstats = []
        t, p = Suite._sig_test_from_data(test, data)

        if test == 'StudentD':
            m = mean(data[0])
            s = std(data[0], ddof=1)
            n = data[0].size
            dstats = [m, s, n]
            es = datagen._eft_sz_studentd(t, n)
        elif test == 'StudentI':
            ms = [mean(x) for x in data]
            ss = [std(x, ddof=1) for x in data]
            ns = [x.size for x in data]
            dstats = ms + ss + ns
            es = datagen._eft_sz_studenti_eq_n(t, ns)
        elif test == 'Welch':
            ms = [mean(x) for x in data]
            ss = [std(x, ddof=1) for x in data]
            ns = [x.size for x in data]
            dstats = ms + ss + ns
            es = datagen._eft_sz_welch(*data)
        elif test == 'ANOVA':
            ms = [mean(x) for x in data]
            ss = [std(x, ddof=1) for x in data]
            ns = [x.size for x in data]
            dstats = ms + ss + ns
            es = datagen._eft_sz_anova(*data)
        elif test == 'MannWhitney':
            ns = [x.size for x in data]
            es = datagen._eft_sz_mann_whitney(p, sum(ns), 2)
        elif test == 'Wilcoxon':
            ns = [x.size for x in data]
            es = datagen._eft_sz_wilcoxon(p, sum(ns), 2)
        elif test == 'KruskalWallis':
            ns = [x.size for x in data]
            es = datagen._eft_sz_kruskall(t, sum(ns), 7)
        elif test == 'Friedman':
            ns = [x.size for x in data]
            es = datagen._eft_sz_kruskall(t, sum(ns), 7)

        return p, es, dstats

if __name__ == "__main__":
    DataGen.gen_datasets(10)
    tests = PARAM + NONPARAM
    for t in tests:
        s = Suite(t, f'data/{t}.txt')
        s.run_suite()
        s.log_results()