from cheno import MannWhitney, Wilcoxon, KruskalWallis, Friedman
from cheno.tests.util import progress_bar
from itertools import zip_longest

from scipy import stats
from numpy.core import numeric as nm

from time import time
import logging
import traceback
import logging
import datetime

# initialize the log settings
logging.basicConfig(filename=f'log/{__file__}_{datetime.datetime.today().strftime("%y%m%d_%H%M")}.log',
                    level=logging.INFO)


def test_mann_whitney_generated_datasets():
    rel_tol = []
    times = []
    print("Testing Mann-Whitney U-test:")
    with open("data/mann_whitney.txt", "r") as f:
        for i, lines in enumerate(_grouper(f, 2, '')):
            progress_bar(i, 10000)
            p = float(lines[0].strip())
            r = float(lines[1].strip())
            tol, ti = rel_tol_mann_whitney(p, r)
            rel_tol.append(tol)
            times.append(ti)
    print(f"\nAccuracy Stats:\n"
          f"---------------\n"
          f"Mean: {nm.mean(rel_tol)},\n"
          f"Min: {min(rel_tol)},\n"
          f"Max: {max(rel_tol)}")
    print(f"\nBenchmark Stats:\n"
          f"---------------\n"
          f"Mean: {nm.mean(times)},\n "
          f"Min: {min(times)},\n"
          f"Max: {max(times)}")


def rel_tol_mann_whitney(p, r):
    m = MannWhitney()
    start = time()
    try:
        ranks, data = m.solve(p, r, [None, None], tails=2)
    except Exception as e:
        logging.error("Error occurred" + str(e))
        logging.error(traceback.format_exc())
    end = time()
    _, p_prime = stats.mannwhitneyu(data[0], data[1])
    return abs((p - p_prime) / p), end - start


def test_wilcoxon_generated_datasets():
    rel_tol = []
    times = []
    print("Testing Wilcoxon t-test:")
    with open("data/wilcoxon.txt", "r") as f:
        for i, lines in enumerate(_grouper(f, 2, '')):
            progress_bar(i, 10000)
            p = float(lines[0].strip())
            r = float(lines[1].strip())
            tol, ti = rel_tol_wilcoxon(p, r)
            rel_tol.append(tol)
            times.append(ti)
    print(f"\nAccuracy Stats:\n"
          f"---------------\n"
          f"Mean: {nm.mean(rel_tol)},\n"
          f"Min: {min(rel_tol)},\n"
          f"Max: {max(rel_tol)}")
    print(f"\nBenchmark Stats:\n"
          f"---------------\n"
          f"Mean: {nm.mean(times)},\n "
          f"Min: {min(times)},\n"
          f"Max: {max(times)}")


def rel_tol_wilcoxon(p, r):
    w = Wilcoxon()
    start = time()
    try:
        ranks, data = w.solve(p, r, tails=2)
    except Exception as e:
        logging.error("Error occurred" + str(e))
        logging.error(traceback.format_exc())
    end = time()
    _, p_prime = stats.wilcoxon(data[0], data[1])
    return abs((p - p_prime) / p), end - start


def test_kruskal_wallis_generated_datasets(k=2, n=100):
    rel_tol = []
    times = []
    print("Testing Kruskal-Wallis H-test:")
    with open(f"data/kruskal_wallis_{k}_{n}.txt", "r") as f:
        for i, lines in enumerate(_grouper(f, 2, '')):
            progress_bar(i, 10000)
            p = float(lines[0].strip())
            eta2 = float(lines[1].strip())
            tol, ti = rel_tol_kruskal_wallis(p, eta2, k)
            rel_tol.append(tol)
            times.append(ti)
    print(f"\nAccuracy Stats:\n"
          f"---------------\n"
          f"Mean: {nm.mean(rel_tol)},\n"
          f"Min: {min(rel_tol)},\n"
          f"Max: {max(rel_tol)}")
    print(f"\nBenchmark Stats:\n"
          f"---------------\n"
          f"Mean: {nm.mean(times)},\n "
          f"Min: {min(times)},\n"
          f"Max: {max(times)}")


def rel_tol_kruskal_wallis(p, eta2, k):
    kw = KruskalWallis(*[None for _ in range(k)])
    start = time()
    try:
        ranks, data = kw.solve(p, eta2)
    except Exception as e:
        logging.error("Error occurred" + str(e))
        logging.error(traceback.format_exc())
    end = time()
    _, p_prime = stats.kruskal(*data)
    return abs((p - p_prime) / p), end - start


def test_friedman_generated_datasets(k=2, n=100):
    rel_tol = []
    times = []
    print("Testing Friedman test:")
    with open(f"data/friedman_{k}_{n}.txt", "r") as f:
        for i, lines in enumerate(_grouper(f, 2, '')):
            progress_bar(i, 10000)
            p = float(lines[0].strip())
            w = float(lines[1].strip())
            tol, ti = rel_tol_friedman(p, w, k)
            rel_tol.append(tol)
            times.append(ti)
    print(f"\nAccuracy Stats:\n"
          f"---------------\n"
          f"Mean: {nm.mean(rel_tol)},\n"
          f"Min: {min(rel_tol)},\n"
          f"Max: {max(rel_tol)}")
    print(f"\nBenchmark Stats:\n"
          f"---------------\n"
          f"Mean: {nm.mean(times)},\n "
          f"Min: {min(times)},\n"
          f"Max: {max(times)}")


def rel_tol_friedman(p, w, k):
    f = Friedman()
    start = time()
    try:
        ranks, data = f.solve(p, w, k)
    except Exception as e:
        logging.error("Error occurred" + str(e))
        logging.error(traceback.format_exc())
    end = time()
    _, p_prime = stats.friedmanchisquare(*data)
    return abs((p - p_prime) / p), end - start


def _grouper(iterable, n, fillvalue=None):
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


if __name__ == "__main__":
    test_mann_whitney_generated_datasets()
    # rel_tol_mann_whitney(0.2450161785357845, -0.11625400197753741)
    test_wilcoxon_generated_datasets()
    test_kruskal_wallis_generated_datasets(k=4, n=100)
    test_kruskal_wallis_generated_datasets(k=7, n=100)
    test_friedman_generated_datasets(k=4, n=100)
    test_friedman_generated_datasets(k=7, n=100)
