from cheno import StudentD, StudentI, Welch, ANOVA
from cheno.tests.test_nonparam_gen import _grouper
from cheno.tests.test_param import _ttest_from_stats
from cheno.tests.util import progress_bar

from time import time
import logging
import traceback
import logging
import datetime

from scipy import stats
from numpy.core import numeric as nm

# initialize the log settings
logging.basicConfig(filename=f'log/{__file__}_{datetime.datetime.today().strftime("%y%m%d_%H%M")}.log', level=logging.INFO)


def test_student_d_generated_datasets():
    rel_tol = []
    times = []
    print("Testing Student's dependent t-test:")
    with open("data/student_d.txt", "r") as f:
        for i, lines in enumerate(_grouper(f, 2, '')):
            progress_bar(i, 10000)
            p = float(lines[0].strip())
            d = float(lines[1].strip())
            tol, ti = rel_tol_student_d(p, d)
            rel_tol.append(tol)
            times.append(ti)
    print(f"\nAccuracy Stats:\n"
          f"---------------\n"
          f"Mean: {nm.mean(rel_tol)},\n"
          f"Min: {min(rel_tol)},\n"
          f"Max: {max(rel_tol)}")
    print(f"\nBenchmark Stats:\n"
          f"---------------\n"
          f"Mean: {nm.mean(times)},\n "
          f"Min: {min(times)},\n"
          f"Max: {max(times)}")


def rel_tol_student_d(p, d):
    s = StudentD(throw_error=False)
    start = time()
    try:
        mean, s, n, data = s.solve(p, d, tails=2)
    except Exception as e:
        logging.error("Error occurred" + str(e))
        logging.error(traceback.format_exc())
    end = time()
    _, p_prime = _ttest_from_stats(mean, s, n)
    return abs((p - p_prime) / p), end - start


def test_student_i_eq_n_generated_datasets():
    rel_tol = []
    times = []
    print("Testing Student's independent eq. sample sizes t-test:")
    with open("data/student_d.txt", "r") as f:
        for i, lines in enumerate(_grouper(f, 2, '')):
            progress_bar(i, 10000)
            p = float(lines[0].strip())
            d = float(lines[1].strip())
            tol, ti = rel_tol_student_i_eq_n(p, d)
            rel_tol.append(tol)
            times.append(ti)
    print(f"\nAccuracy Stats:\n"
          f"---------------\n"
          f"Mean: {nm.mean(rel_tol)},\n"
          f"Min: {min(rel_tol)},\n"
          f"Max: {max(rel_tol)}")
    print(f"\nBenchmark Stats:\n"
          f"---------------\n"
          f"Mean: {nm.mean(times)},\n "
          f"Min: {min(times)},\n"
          f"Max: {max(times)}")


def rel_tol_student_i_eq_n(p, d):
    s = StudentI()
    start = time()
    try:
        means, ss, ns, data = s.solve(p, d, tails=2, eq_n=True)
    except Exception as e:
        logging.error("Error occurred" + str(e))
        logging.error(traceback.format_exc())
    end = time()
    _, p_prime = stats.ttest_ind(data[0], data[1])
    return abs((p - p_prime) / p), end - start


def test_student_i_uneq_n_generated_datasets():
    rel_tol = []
    times = []
    print("Testing Student's independent uneq. sample sizes t-test:")
    with open("data/student_d.txt", "r") as f:
        for i, lines in enumerate(_grouper(f, 2, '')):
            progress_bar(i, 10000)
            p = float(lines[0].strip())
            d = float(lines[1].strip())
            tol, ti = rel_tol_student_i_uneq_n(p, d)
            rel_tol.append(tol)
            times.append(ti)
    print(f"\nAccuracy Stats:\n"
          f"---------------\n"
          f"Mean: {nm.mean(rel_tol)},\n"
          f"Min: {min(rel_tol)},\n"
          f"Max: {max(rel_tol)}")
    print(f"\nBenchmark Stats:\n"
          f"---------------\n"
          f"Mean: {nm.mean(times)},\n "
          f"Min: {min(times)},\n"
          f"Max: {max(times)}")


def rel_tol_student_i_uneq_n(p, d):
    s = StudentI()
    start = time()
    try:
        means, ss, ns, data = s.solve(p, d, tails=2, eq_n=False)
    except Exception as e:
        logging.error("Error occurred" + str(e))
        logging.error(traceback.format_exc())
    end = time()
    _, p_prime = stats.ttest_ind(data[0], data[1])
    return abs((p - p_prime) / p), end - start


def test_welch_generated_datasets():
    rel_tol = []
    times = []
    print("Testing Welchs's t-test:")
    with open("data/welch.txt", "r") as f:
        for i, lines in enumerate(_grouper(f, 2, '')):
            progress_bar(i, 10000)
            p = float(lines[0].strip())
            d = float(lines[1].strip())
            tol, ti = rel_tol_welch(p, d)
            rel_tol.append(tol)
            times.append(ti)
    print(f"\nAccuracy Stats:\n"
          f"---------------\n"
          f"Mean: {nm.mean(rel_tol)},\n"
          f"Min: {min(rel_tol)},\n"
          f"Max: {max(rel_tol)}")
    print(f"\nBenchmark Stats:\n"
          f"---------------\n"
          f"Mean: {nm.mean(times)},\n "
          f"Min: {min(times)},\n"
          f"Max: {max(times)}")


def rel_tol_welch(p, d):
    w = Welch()
    start = time()
    try:
        m, s, n, data = w.solve(p, d)
    except Exception as e:
        logging.error("Error occurred" + str(e))
        logging.error(traceback.format_exc())
    end = time()
    _, p_prime = stats.ttest_ind(data[0], data[1], equal_var=False)
    return abs((p - p_prime) / p), end - start


def test_anova_eq_n_generated_datasets(k, n):
    rel_tol = []
    times = []
    print("Testing ANOVA:")
    with open(f"data/anova_eq_n_{k}_{n}.txt", "r") as f:
        for i, lines in enumerate(_grouper(f, 2, '')):
            progress_bar(i, 10000)
            p = float(lines[0].strip())
            eta2 = float(lines[1].strip())
            tol, ti = measure_anova_eq_n(p, eta2, k)
            rel_tol.append(tol)
            times.append(ti)
    print(f"\nAccuracy Stats:\n"
          f"---------------\n"
          f"Mean: {nm.mean(rel_tol)},\n"
          f"Min: {min(rel_tol)},\n"
          f"Max: {max(rel_tol)}")
    print(f"\nBenchmark Stats:\n"
          f"---------------\n"
          f"Mean: {nm.mean(times)},\n "
          f"Min: {min(times)},\n"
          f"Max: {max(times)}")


def measure_anova_eq_n(p, eta2, k):
    a = ANOVA(k)
    start = time()
    try:
        means, ss, ns, data = a.solve(p, eta2)
    except Exception as e:
        logging.error("Error occurred" + str(e))
        logging.error(traceback.format_exc())
    end = time()
    _, p_prime = stats.f_oneway(*data)
    return abs((p - p_prime) / p), end - start


if __name__ == "__main__":
    # test_student_d_generated_datasets()
    # test_student_i_eq_n_generated_datasets()
    # test_student_i_uneq_n_generated_datasets()
    # test_anova_eq_n_generated_datasets(4, 100)
    test_anova_eq_n_generated_datasets(7, 100)
    test_welch_generated_datasets()
