import cheno

import sys
from math import sqrt

from numpy import mean, random, std
from scipy import stats


def progress_bar(value, endvalue, bar_length=20):
    percent = float(value) / endvalue
    arrow = '-' * int(round(percent * bar_length) - 1) + '>'
    spaces = ' ' * (bar_length - len(arrow))
    sys.stdout.write("\rPercent: [{0}] {1}%".format(arrow + spaces, int(round(percent * 100))))
    sys.stdout.flush()


def _sample_data_for_test(test):
    if test == 'StudentD':
        t = cheno.StudentD()
        m = random.uniform(0, 2)
        s = random.uniform(t.low_s, t.high_s)
        n = random.randint(t.low_n, t.high_n + 1)
        data = random.normal(m, s, n)
        return [data]
    elif test == 'StudentI':
        t = cheno.StudentI()
        ms = [random.uniform(m[0], m[1]) for m in [[0, 5], [0, 5]]]
        s1 = random.uniform(1, 5)
        ss = [s1, s1 * random.uniform(.9, 1.1)]
        ns = [random.randint(m[0], m[1] + 1) for m in [t.n1, t.n2]]
        return [random.normal(m, s, n) for m, s, n in zip(ms, ss, ns)]
    elif test == 'Welch':
        t = cheno.Welch()
        ms = [random.uniform(m[0], m[1]) for m in [[0, 5], [0, 5]]]
        ss = [random.uniform(s[0], s[1]) for s in [[1, 5], [1, 5]]]
        ns = [random.randint(m[0], m[1] + 1) for m in [t.n1, t.n2]]
        return [random.normal(m, s, n) for m, s, n in zip(ms, ss, ns)]
    elif test == 'ANOVA':
        t = cheno.ANOVA(k=7)
        m1 = random.uniform(0, 5)
        ms = [m1 * random.uniform(.5, 1.5) for _ in range(7)]
        s1 = random.uniform(1, 5)
        ss = [s1 * random.uniform(.9, 1.1) for _ in range(7)]
        ns = [random.randint(m[0], m[1] + 1) for m in t.n_bounds]
        return [random.normal(m, s, n) for m, s, n in zip(ms, ss, ns)]
    elif test == 'MannWhitney':
        t = cheno.MannWhitney()
        ns = [random.randint(m[0], m[1] + 1) for m in [(t.low_n, t.high_n), (t.low_n, t.high_n)]]
        return [random.sample(n) for n in ns]
    elif test == 'Wilcoxon':
        t = cheno.Wilcoxon()
        n = random.randint(t.lb, t.ub + 1)
        return [random.sample(n) for _ in range(2)]
    elif test == 'KruskalWallis':
        t = cheno.ANOVA(k=7)
        ns = [random.randint(8, m[1] + 1) for m in t.n_bounds]
        return [random.sample(n) for n in ns]
    elif test == 'Friedman':
        t = cheno.ANOVA(k=7)
        n = random.randint(8, t.n_bounds[0][1] + 1)
        return [random.sample(n) for _ in range(7)]


def _d_ttest_from_stats(mean, s, n):
    t = mean / (s / sqrt(n))
    p = stats.t.sf(abs(t), n - 1) * 2
    return t, p


def _mann_whitney_from_stats(ranks):
    s1, s2 = sum(ranks[0]), sum(ranks[1])
    w = s1 if s1 < s2 else s2
    n1, n2 = len(ranks[0]), len(ranks[1])
    w_mean = n1 * (n1 + n2 + 1) / 2
    se_w = sqrt(n1 * n2 * (n1 + n2 + 1) / 12)
    z = (w - w_mean) / se_w
    return stats.norm.cdf(z) * 2


def _wilcoxon_from_stats(ranks):
    s1, s2 = sum(ranks[0]), sum(ranks[1])
    w = s1 if s1 < s2 else s2
    n1, n2 = len(ranks[0]), len(ranks[1])
    w_mean = min(n1, n2) * (n1 + n2 + 1) / 2
    se_w = sqrt(n1 * n2 * (n1 + n2 + 1) / 12)
    z = (w - w_mean) / se_w
    return stats.norm.cdf(z) * 2


def _kruskal_wallis_from_stats(ranks):
    s = [(sum(x), len(x)) for x in ranks]
    n = sum([x[1] for x in s])
    h = (12 / (n * (n + 1))) * sum([x[0] ** 2 / x[1] for x in s]) - 3 * (n + 1)
    return stats.chi2.sf(h, 3)


def _friedman_from_stats(ranks):
    k = len(ranks.T)
    s = [(sum(x), len(x)) for x in ranks.T]
    n = sum([x[1] for x in s]) / k
    h = (12 / (n * k * (k + 1))) * sum([x[0] ** 2 for x in s]) - 3 * n * (k + 1)
    return stats.chi2.sf(h, k - 1)


def _eft_sz_studentd(t, n):
    return t / sqrt(n)


def _eft_sz_studenti_eq_n(t, ns):
    return t * sqrt(1 / ns[0] + 1 / ns[1])


def _eft_sz_welch(x, y):
    m1, s1, n1 = mean(x), std(x, ddof=1), len(x)
    m2, s2, n2 = mean(y), std(y, ddof=1), len(y)
    dof = (n1 + n2 - 2)
    d = (m1 - m2) / sqrt(((n1 - 1) * s1 ** 2 + (n2 - 1) * s2 ** 2) / dof)
    return d


def _eft_sz_anova(*args):
    ns = []
    means = []
    sds = []
    for x in args:
        ns.append(len(x))
        means.append(mean(x))
        sds.append(std(x, ddof=1))
    g_mean = sum([m * n for m, n in zip(means, ns)]) / sum(ns)
    df_m = len(args) - 1
    df_r = sum(ns) - len(args)
    ss_m = sum([n * (m - g_mean) ** 2 for m, n in zip(means, ns)])
    ss_r = sum((n - 1) * s ** 2 for s, n in zip(sds, ns))
    return ss_m / (ss_m + ss_r)


def _eft_sz_mann_whitney(p, n, tails=1):
    z = stats.norm.ppf(p / tails)
    return z / sqrt(n)


def _eft_sz_wilcoxon(p, n, tails=1):
    z = stats.norm.ppf(p / tails)
    return z / sqrt(n)


def _eft_sz_kruskall(h, n, k):
    return (h - k + 1) / (n - k)


def _eft_sz_friedman(f, n, k):
    return f / (n * (k - 1))
