import warnings
import functools
import random
from math import sqrt
from numpy.random import normal, uniform
import numpy as np

def deprecated(func):
    '''This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.'''

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn('Call to deprecated function {}.'.format(func.__name__),
                      category=DeprecationWarning,
                      stacklevel=2)
        warnings.simplefilter('default', DeprecationWarning)  # reset filter
        return func(*args, **kwargs)

    return new_func


def count_none(a):
    return len([1 for x in a if x is None])


def count_non_null(a):
    return len([1 for x in a if x is not None])


def get_negative(a):
    return [x for x in a if x is not None and type(x) is not str and x < 0]


def get_not_nan(a):
    return [x for x in a if x is not None]


def get_not_nan_idx(a):
    return [i for i, x in enumerate(a) if x is not None]


def get_nan_idx(a):
    return [i for i, x in enumerate(a) if x is None]


def rand_generate_subset(SZ1, SZ2, S1):
    N = SZ1 + SZ2
    lower_bound = (SZ1 * (SZ1 + 1)) / 2
    upper_bound = SZ1 * (N + (N - SZ1 + 1)) / 2

    if S1 < lower_bound or S1 > upper_bound:
        raise ValueError(f'The sum you have inputed has to be between: {lower_bound} and {upper_bound}')

    array = range(1, N + 1)

    # Init solution
    sol = sorted(random.sample(array, SZ1))
    rest = [x for x in array if x not in sol]
    if sum(sol) == S1:
        return sol, rest

    # Set best solution
    min_d = abs(S1 - sum(sol))
    best_sol = sol

    # Local improvements
    for i in range(0, N ** 2):
        if S1 - sum(sol) < 0:
            next_lesser, next_lesser_idx = _next_lesser(rest, sol[SZ1 - (i % SZ1) - 1])
            if next_lesser != -1:
                rest[next_lesser_idx] = sol[SZ1 - (i % SZ1) - 1]
                sol[SZ1 - (i % SZ1) - 1] = next_lesser
            else:
                sol = sorted(sol)
        # TODO: Think about doing it symetrically. What I mean by that is that in the above case
        # you pick from the end of the array(right part) and move them to the beginning (left part), 
        # but also in the below case. Maybe there are better ways to move them around,  for example 
        # the above case, starting from the left, and in this case starting from the right.
        elif S1 - sum(sol) > 0:
            next_greater, next_greater_idx = _next_greater(rest, sol[SZ1 - (i % SZ1) - 1])
            if next_greater != -1:
                rest[next_greater_idx] = sol[SZ1 - (i % SZ1) - 1]
                sol[SZ1 - (i % SZ1) - 1] = next_greater
            else:
                sol = sorted(sol)
        else:
            return [sol, [x for x in array if x not in sol]]

        if abs(S1 - sum(sol)) < min_d:
            best_sol = sol

    return [sol, [x for x in array if x not in sol]]


def _gen_param(means, stds, ns):
    if len(means) != len(stds) or len(stds) != len(ns):
        raise ValueError(f'Unequal size of parameters.')
    datasets = []
    for m, s, n in zip(means, stds, ns):
        s = sqrt(s ** 2 * ((n - 1) / n))
        datasets.append(_adjust(normal(m, s, n), m, s))
    return datasets


def _gen_nonparam(ranks, n, low=-5, high=5, dist='uniform'):
    data = sorted(uniform(low, high, n))
    if dist == 'normal':
        raise NotImplementedError("Normal distribution generation is not yet implemented.")

    res = []
    for idx in ranks:
        res.append([data[i - 1] for i in idx])

    return res


def _adjust(data, mean, sd):
    return _adj_mean(_adj_std(data, target=sd), target=mean)


def _adj_mean(data, target=None, inc=None):
    y = data.mean()
    d = 0
    if target is not None:
        d = y - target
    elif inc is not None:
        d = inc
    else:
        raise ValueError('No manipulation strategy chosen.')
    return data - d


def _adj_std(data, target=None, inc=None):
    y = data.std()
    d = 1
    if target is not None:
        d = target / y
    elif inc is not None:
        d = inc
    else:
        raise ValueError('No manipulation strategy chosen.')
    return data * d


def _next_lesser(rest, v):
    if rest is None or len(rest) <= 0:
        raise ValueError('Comparing values in an empty list.')
    if not all(rest[i] <= rest[i + 1] for i in range(len(rest) - 1)):
        raise ValueError('List is not sorted.')

    for i, e in enumerate(reversed(rest)):
        if v > e:
            return e, len(rest) - i - 1
    return -1, -1


def _next_greater(rest, v):
    if rest is None or len(rest) <= 0:
        raise ValueError('Comparing values in an empty list.')
    if not all(rest[i] <= rest[i + 1] for i in range(len(rest) - 1)):
        raise ValueError('List is not sorted.')

    for i, e in enumerate(rest):
        if v < e:
            return e, i
    return -1, -1


def cohan_d(n1, n2, v1, v2, m1, m2):
    m_diff = m1 - m2
    df1 = n1 - 1
    df2 = n2 - 1
    denom = sqrt((df1 * v1 + df2 * v2) / (df1 + df2))
    return m_diff / denom


if __name__ == '__main__':
    # rand_generate_subset(28, 72, 2414)
    # x = cohan_d(10, 20, 9, 0.9, 20.6, 22.1)
    # x = _gen_param([5.50], [sqrt(9.67)], [4])
    # x = _gen_param([5.72639934], [sqrt(10.559349299964584)], [3])
    # x = _gen_param([3.02452527], [sqrt(12.448104350948586)], [3])
    # x = _gen_param([4.50], [sqrt(11.67)], [4])
    # x = _gen_param([6.25], [sqrt(4.25)], [4])
    # x = _gen_param([6.25], [sqrt(0.92)], [4])
    # x = _gen_param([5.25], [sqrt(15.58)], [4])
    # x = _gen_param([6.75], [sqrt(20.92)], [4])

    x = _gen_param([4.13], [sqrt(8.25)], [8])
    print(x)
