import multiprocessing as mp
from itertools import product
from math import isclose, sqrt
from random import uniform

import numpy as np
from scipy import stats

from cheno.util import _gen_param, count_none


class Welch(object):

    def __init__(self, n1=[5, 100], n2=[5, 100], s1=[.1, 2], s2=[.1, 2], m1=[0, 7], m2=[0, 7], throw_error=True):
        self.n1 = n1
        self.n2 = n2
        self.s1 = s1
        self.s2 = s2
        self.m1 = m1
        self.m2 = m2
        self.throw_error = throw_error

    def solve(self, p, d, ns=None, means=None, sds=None, tails=2, splits=10):
        if tails != 1 and tails != 2:
            if self.throw_error:
                raise ValueError(
                    "Invalid number of tails. NHST accept either one tail or two tails.")
            else:
                tails = 2
        ns, means, sds = self._init_parameters(ns, means, sds)

        ns, sds = self._find_sample_szs_sds(p, d, ns, sds, tails, splits)

        means = self._find_means(p, ns, means, sds, tails)

        return means, sds, ns, _gen_param(means, sds, ns)

    def _init_parameters(self, ns, means, sds):
        means = [None, None] if means is None else means
        sds = [None, None] if sds is None else sds
        ns = [None, None] if ns is None else ns

        # General check on the size of the constraints
        if len(ns) != 2:
            if self.throw_error:
                raise ValueError(
                    "Provided number of sample size doesn't match number of groups.")
            else:
                ns = [None, None]
        if len(means) != 2:
            if self.throw_error:
                raise ValueError(
                    "Provided number of means doesn't match number of groups")
            else:
                means = [None, None]
        if len(sds) != 2:
            if self.throw_error:
                raise ValueError(
                    "Provided number of standard deviations doesn't match number of groups")
            else:
                sds = [None, None]

        return ns, means, sds

    def _find_sample_szs_sds(self, p, d, ns, sds, tails, splits):

        _, _, _, s1, s2 = self._find_nearest_comb(p, d, ns, sds, tails, splits)

        combs = [[p], [d], [ns[0]], [ns[1]], [s1], [s2], [tails]]

        # Single user sample size provided:
        if count_none(ns) == 1:
            idx_prov_n, prov_n = (
                [(i, x) for i, x in enumerate(ns) if x is not None])[0]
            combs = product([p], [d], [prov_n], range(
                self.n2[0], self.n2[1]), [s1], [s2], [tails])
            if idx_prov_n == 1:
                combs = product([p], [d], range(self.n1[0], self.n1[1]), [
                                prov_n], [s1], [s2], [tails])

        # No user sample sizes provided
        if count_none(ns) == 2:
            combs = product([p], [d], range(self.n1[0], self.n1[1] + 1), range(self.n2[0], self.n2[1]),
                            [s1], [s2], [tails])

        with mp.Pool(processes=5) as pool:
            results = pool.starmap(self._diff_t_from_p_d, combs)

        best_comb = min(results, key=lambda t: t[0])

        return best_comb[1:3], best_comb[3:5]

    def _find_means(self, p, ns, means, sds, tails):
        vn1 = sds[0] ** 2 / ns[0]
        vn2 = sds[1] ** 2 / ns[1]

        with np.errstate(divide='ignore', invalid='ignore'):
            df = (vn1 + vn2) ** 2 / (vn1 ** 2 /
                                     (ns[0] - 1) + vn2 ** 2 / (ns[0] - 1))

        # If df is undefined, variances are zero (assumes n1 > 0 & n2 > 0).
        # Hence it doesn't matter what df is as long as it's not NaN.
        df = np.where(np.isnan(df), 1, df)
        t = abs(stats.t.ppf(1 - p / tails, df))
        mean_diff = t * sqrt(vn1 + vn2)

        if count_none(means) == 0 and isclose(means[0] - means[1], mean_diff, abs_tol=1):
            if self.throw_error:
                raise ValueError("Provided means will not produce desired inference statistics. "
                                 "Pick different means or remove them altogether.")
        if count_none(means) == 1:
            prov_mean_idx, prov_mean = (
                [(i, x) for i, x in enumerate(means) if x is not None])[0]
            means[1 - prov_mean_idx] = mean_diff + \
                prov_mean if prov_mean_idx == 1 else prov_mean - mean_diff
        if count_none(means) == 2:
            means[0] = uniform(self.m1[0], self.m2[0])
            means[1] = means[0] - mean_diff

        return means

    def _find_nearest_comb(self, p, d, ns, sds, tails, s=10):

        n1_range = [ns[0]]
        if ns[0] is None:
            n1_range = [(self.n1[0] + i * (self.n1[1] - self.n1[0]) / s)
                        for i in range(s - 1)]
        n2_range = [ns[1]]
        if ns[1] is None:
            n2_range = [(self.n2[0] + i * (self.n2[1] - self.n2[0]) / s)
                        for i in range(s - 1)]
        s1_range = [sds[0]]
        if sds[0] is None:
            s1_range = [(self.s1[0] + i * (self.s1[1] - self.s1[0]) / s)
                        for i in range(s - 1)]
        s2_range = [sds[1]]
        if sds[1] is None:
            s2_range = [(self.s2[0] + i * (self.s2[1] - self.s2[0]) / s)
                        for i in range(s - 1)]

        combs = product([p], [d], n1_range, n2_range,
                        s1_range, s2_range, [tails])
        with mp.Pool(processes=5) as pool:
            results = pool.starmap(self._diff_t_from_p_d, combs)

        return min(results, key=lambda t: t[0])

    @staticmethod
    def _diff_t_from_p_d(p, d, n1, n2, s1, s2, tails):
        df1 = n1 - 1
        df2 = n2 - 1
        vn1 = s1 ** 2 / n1
        vn2 = s2 ** 2 / n2

        with np.errstate(divide='ignore', invalid='ignore'):
            df = (vn1 + vn2) ** 2 / (vn1 ** 2 / (n1 - 1) + vn2 ** 2 / (n2 - 1))

        # If df is undefined, variances are zero (assumes n1 > 0 & n2 > 0).
        # Hence it doesn't matter what df is as long as it's not NaN.
        df = np.where(np.isnan(df), 1, df)

        t_from_p = abs(stats.t.ppf(1 - p / tails, df))
        t_from_d = abs(d * sqrt((df1 * s1 ** 2 + df2 * s2 **
                                 2) / (df1 + df2)) / sqrt(vn1 + vn2))

        return abs(t_from_p - t_from_d), n1, n2, s1, s2


if __name__ == "__main__":
    # Example
    p = 0.1541255195640466
    d = -0.8013749652081635

    # Commands
    w = Welch(s1=[.1, 10], s2=[.1, 10])
    m, s, n, _ = w.solve(p, d)

    # Results
    print(stats.ttest_ind_from_stats(
        m[0], s[0], n[0], m[1], s[1], n[1], equal_var=False))
